from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input


def xgboost_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Estimators: The number of trees in the forest.")
    print("Please specify the number of trees in the forest. A good starting range could be between 50 and 500, such as 100.")
    n_estimators = num_input(SECTION[2], "@N Estimators: ")
    print("Learning Rate: It controls the step-size in updating the weights. It shrinks the contribution of each tree by learning_rate. There is a trade-off between learning_rate and n_estimators.")
    print("Please specify the initial learning rate of Xgboost, such as 0.1.")
    learning_rate = float_input(0.01, SECTION[2], "@Learning Rate: ")
    print("Max Depth: The maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit.")
    print("Please specify the maximum depth of a tree. A good starting range could be between 3 and 15, such as 4.")
    max_depth = num_input(SECTION[2], "@Max Depth: ")
    print("Subsample: The fraction of samples to be used for fitting the individual base learners.")
    print("Please specify the fraction of samples to be used for fitting the individual base learners. A good starting range could be between 0.5 and 1.0, such as 0.8.")
    subsample = float_input(1, SECTION[2], "@Subsample: ")
    print("Colsample Bytree: The fraction of features to be used for fitting the individual base learners.")
    print("Please specify the fraction of features to be used for fitting the individual base learners. A good starting range could be between 0.5 and 1.0, such as 1.")
    colsample_bytree = float_input(1, SECTION[2], "@Colsample Bytree: ")
    print("Alpha: L1 regularization term on weights.")
    print("Please specify the L1 regularization term on weights. A good starting range could be between 0 and 1.0, such as 0.")
    alpha = float_input(0, SECTION[2], "@Alpha: ")
    print("Lambda: L2 regularization term on weights.")
    print("Please specify the L2 regularization term on weights. A good starting range could be between 0 and 1.0, such as 1.")
    lambd = float_input(1, SECTION[2], "@Lambda: ")
    hyper_parameters = {
        "n_estimators": n_estimators,
        "learning_rate": learning_rate,
        "max_depth": max_depth,
        "subsample": subsample,
        "colsample_bytree": colsample_bytree,
        "alpha": alpha,
        "lambd": lambd,
    }
    return hyper_parameters


# def feature_importance_value(data: pd.DataFrame, trained_model: any) -> None:
#     """
#     Draw the feature importance value orderly for analysis.

#     Parameters
#     ----------
#     data: pd.DataFrame (n_samples, n_components)
#         Data for silhouette.

#     trained_model: any
#         The algorithm which to be used

#     algorithm_name : str
#         the name of the algorithm

#     References
#     ----------
#     XGBoost is an optimized distributed gradient boosting library designed to be highly efficient, f
#     lexible and portable. It implements machine learning algorithms under the Gradient Boosting fram
#     ework. XGBoost provides a parallel tree boosting (also known as GBDT, GBM) that solve many data
#     science problems in a fast and accurate way.

#     https://xgboost.readthedocs.io/en/stable/
#     """
#     columns_name = data.columns
#     for feature_name, score in zip(list(columns_name), trained_model.feature_importances_):
#         print(feature_name, ":", score)


# def feature_weights_histograms(trained_model: any, image_config: dict, algorithm_name: str) -> None:
#     """
#     Draw the histograms of feature weights for XGBoost predictions.

#     Parameters
#     ----------
#     data: pd.DataFrame (n_samples, n_components)
#         Data for silhouette.

#     trained_model: any
#         The algorithm which to be used

#     algorithm_name : str
#         the name of the algorithm

#     References
#     ----------
#     XGBoost is an optimized distributed gradient boosting library designed to be highly efficient, f
#     lexible and portable. It implements machine learning algorithms under the Gradient Boosting fram
#     ework. XGBoost provides a parallel tree boosting (also known as GBDT, GBM) that solve many data
#     science problems in a fast and accurate way.

#     https://xgboost.readthedocs.io/en/stable/
#     """

#     # create drawing canvas
#     fig, ax = plt.subplots(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])

#     ax.bar(
#         image_config["bar_x"],
#         trained_model.feature_importances_,
#         tick_label=image_config["bar_label"],
#         width=image_config["bar_width"],
#         bottom=image_config["bottom"],
#     )

#     # automatically optimize picture layout structure
#     fig.tight_layout()
#     xmin, xmax = ax.get_xlim()
#     ymin, ymax = ax.get_ylim()
#     x_adjustment = (xmax - xmin) * 0.1
#     y_adjustment = (ymax - ymin) * 0.1
#     ax.axis([xmin - x_adjustment, xmax + x_adjustment, ymin - y_adjustment, ymax + y_adjustment])

#     # convert the font of the axes
#     plt.tick_params(labelsize=image_config["labelsize"])  # adjust the font size of the axis label
#     # plt.setp(ax.get_xticklabels(), rotation=image_config['xrotation'], ha=image_config['xha'],
#     #          rotation_mode="anchor")  # axis label rotation Angle
#     # plt.setp(ax.get_yticklabels(), rotation=image_config['rot'], ha=image_config['yha'],
#     #          rotation_mode="anchor")  # axis label rotation Angle
#     x1_label = ax.get_xticklabels()  # adjust the axis label font
#     [x1_label_temp.set_fontname(image_config["axislabelfont"]) for x1_label_temp in x1_label]
#     y1_label = ax.get_yticklabels()
#     [y1_label_temp.set_fontname(image_config["axislabelfont"]) for y1_label_temp in y1_label]

#     ax.set_title(
#         label=algorithm_name,
#         fontdict={
#             "size": image_config["title_size"],
#             "color": image_config["title_color"],
#             "family": image_config["title_font"],
#         },
#         loc=image_config["title_location"],
#         pad=image_config["title_pad"],
#     )


# def feature_importance_map(trained_model: any, image_config: dict, algorithm_name: str) -> None:
#     """
#     Draw the diagram of feature importance map ranked by importance for analysis.

#     Parameters
#     ----------
#     trained_model: any
#         The algorithm which to be used

#     algorithm_name : str
#         the name of the algorithm

#     References
#     ----------
#     XGBoost is an optimized distributed gradient boosting library designed to be highly efficient, f
#     lexible and portable. It implements machine learning algorithms under the Gradient Boosting fram
#     ework. XGBoost provides a parallel tree boosting (also known as GBDT, GBM) that solve many data
#     science problems in a fast and accurate way.

#     https://xgboost.readthedocs.io/en/stable/
#     """
#     # create drawing canvas
#     fig, ax = plt.subplots(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])

#     # draw the main content
#     xgboost.plot_importance(booster=trained_model, ax=image_config["ax"], height=image_config["bar_width"])

#     # automatically optimize picture layout structure
#     fig.tight_layout()
#     xmin, xmax = ax.get_xlim()
#     ymin, ymax = ax.get_ylim()
#     x_adjustment = (xmax - xmin) * 0.1
#     y_adjustment = (ymax - ymin) * 0.1
#     ax.axis([xmin - x_adjustment, xmax + x_adjustment, ymin - y_adjustment, ymax + y_adjustment])

#     # convert the font of the axes
#     plt.tick_params(labelsize=image_config["labelsize"])  # adjust the font size of the axis label
#     # plt.setp(ax.get_xticklabels(), rotation=image_config['xrotation'], ha=image_config['xha'],
#     #          rotation_mode="anchor")  # axis label rotation Angle
#     # plt.setp(ax.get_yticklabels(), rotation=image_config['rot'], ha=image_config['yha'],
#     #          rotation_mode="anchor")  # axis label rotation Angle
#     x1_label = ax.get_xticklabels()  # adjust the axis label font
#     [x1_label_temp.set_fontname(image_config["axislabelfont"]) for x1_label_temp in x1_label]
#     y1_label = ax.get_yticklabels()
#     [y1_label_temp.set_fontname(image_config["axislabelfont"]) for y1_label_temp in y1_label]

#     ax.set_title(
#         label=algorithm_name,
#         fontdict={
#             "size": image_config["title_size"],
#             "color": image_config["title_color"],
#             "family": image_config["title_font"],
#         },
#         loc=image_config["title_location"],
#         pad=image_config["title_pad"],
#     )



import copy
import json
import os
from typing import Dict, Optional, Tuple

import mlflow
import pandas as pd
from imblearn.over_sampling import RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler
from sklearn.feature_selection import GenericUnivariateSelect, SelectKBest
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import MinMaxScaler, PolynomialFeatures, StandardScaler

from ..constants import MLFLOW_ARTIFACT_DATA_PATH
from ..utils.base import save_data, save_model, save_text
from .data_readiness import np2pd


class PipelineConstrutor:
    """Construct a sklearn pipeline from a dictionary of transformers."""

    @property
    def transformer_dict(self) -> Dict:
        """A dictionary of transformers. Need to be updated when new transformers in the customized automated ML pipeline is added."""
        return {
            "SimpleImputer": SimpleImputer,
            "MinMaxScaler": MinMaxScaler,
            "StandardScaler": StandardScaler,
            "PolynomialFeatures": PolynomialFeatures,
            "RandomOverSampler": RandomOverSampler,
            "RandomUnderSampler": RandomUnderSampler,
            "GenericUnivariateSelect": GenericUnivariateSelect,
            "SelectKBest": SelectKBest,
        }

    def chain(self, transformer_config: Dict) -> object:
        """Chain transformers together into a sklearn pipeline.

        Parameters
        ----------
        transformer_config : Dict
            A dictionary of transformers and their parameters.

        Returns
        -------
        object
            A sklearn pipeline.
        """
        transformers = []
        for transformer_name, transformer_params in transformer_config.items():
            transformers.append(self.transformer_dict[transformer_name](**transformer_params))
        return make_pipeline(*transformers)


def build_transform_pipeline(imputation_config: Dict, feature_scaling_config: Dict, feature_selection_config: Dict, run: object, X_train: pd.DataFrame, y_train: pd.DataFrame) -> Tuple[Dict, object]:
    """Build the transform pipeline.

    Parameters
    ----------
    imputation_config : Dict
        The imputation configuration.

    feature_scaling_config : Dict
        The feature scaling configuration.

    feature_selection_config : Dict
        The feature selection configuration.

    run : object
        The model selection object.

    X_train : pd.DataFrame
        The training data.

    Returns
    -------
    Tuple[Dict, object]
        The transform pipeline configuration and the transform pipeline object.
    """
    print("Build the transform pipeline according to the previous operations.")
    # Aggregate transformer configuartion.
    transformer_config = {}
    transformer_config.update(imputation_config)
    transformer_config.update(feature_scaling_config)
    transformer_config.update(feature_selection_config)
    transformer_config.update(run.transformer_config)
    transformer_config_str = copy.deepcopy(transformer_config)
    for key, value in transformer_config_str.items():
        for k, v in value.items():
            if callable(v):
                transformer_config_str[key][k] = v.__name__
    transformer_config_str = json.dumps(transformer_config_str, indent=4)
    GEOPI_OUTPUT_ARTIFACTS_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_PATH")
    save_text(transformer_config_str, "Transform Pipeline Configuration", GEOPI_OUTPUT_ARTIFACTS_PATH, "root")
    # If transformer_config is not {}, then create the transform pipeline.
    if transformer_config:
        # Create the transform pipeline.
        transform_pipeline = PipelineConstrutor().chain(transformer_config)
        # Fit the transform pipeline with the training data.
        transform_pipeline.fit(X_train, y_train)
        # Save the transform pipeline.
        GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH")
        save_model(transform_pipeline, "Transform Pipeline", X_train.iloc[[0]], GEOPI_OUTPUT_ARTIFACTS_MODEL_PATH)
    else:
        transform_pipeline = None
    return transformer_config, transform_pipeline


def model_inference(inference_data: pd.DataFrame, is_inference: bool, run: object, transformer_config: Dict, transform_pipeline: Optional[object] = None):
    """Run the model inference.

    Parameters
    ----------
    inference_data : pd.DataFrame
        The inference data.

    is_inference : bool
        Whether to run the model inference.

    run : object
        The model selection object.

    transformer_config : Dict
        The transformer configuration.

    transform_pipeline : Optional[object], optional
        The transform pipeline object. The default is None.
    """
    # If is_inference is True, then run the model inference.
    if is_inference is True:
        print("Use the trained model to make predictions on the application data.")
        # If transformer_config is not {}, then transform the inference data with the transform pipeline.
        if transformer_config:
            inference_data_transformed = transform_pipeline.transform(inference_data)
        else:
            inference_data_transformed = inference_data
        loaded_model = mlflow.sklearn.load_model(f"runs:/{mlflow.active_run().info.run_id}/{run.model_name}")
        inference_data_predicted_np = loaded_model.predict(inference_data_transformed)
        inference_data_predicted = np2pd(inference_data_predicted_np, ["Predicted Value"])
        GEOPI_OUTPUT_ARTIFACTS_DATA_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH")
        save_data(inference_data_predicted, "Application Data Predicted", GEOPI_OUTPUT_ARTIFACTS_DATA_PATH, MLFLOW_ARTIFACT_DATA_PATH)



# -*- coding: utf-8 -*-
import os
import string
import time
from typing import Dict

import pandas as pd
from rich import print

from ..constants import MLFLOW_ARTIFACT_DATA_PATH, OPTION, SECTION
from ..plot.statistic_plot import basic_statistic
from ..utils.base import clear_output, save_data
from .data_readiness import basic_info, limit_num_input, num2option, num_input, show_data_columns


class FeatureConstructor(object):
    """Construct new feature based on the existing data set."""

    oper = "+-*/^(),."
    # parenthesis = ['(', ')']
    cal_words = ["pow", "sin", "cos", "tan", "pi", "mean", "std", "var", "log"]

    def __init__(self, data: pd.DataFrame) -> None:
        self.feature_name = None
        self.data = data
        self.alphabet = string.ascii_lowercase
        self._infix_expr = []
        self._postfix_expr = []
        self.map_dict = {}
        self._result = None
        self.config = {}

    def index2name(self) -> None:
        """Show the index of columns in the data set. The display pattern is [letter : column name], e.g. a : 1st column name; b : 2nd column name."""
        columns_name = self.data.columns
        print("Selected data set:")
        for i in range(len(columns_name)):
            print(self.alphabet[i] + " - " + columns_name[i])
            self.map_dict[self.alphabet[i]] = columns_name[i]

    def _get_column(self, index: str) -> str:
        return self.map_dict[index]

    def input_feature_name(self) -> None:
        """Name the constructed feature (column name), like 'NEW-COMPOUND'."""
        while True:
            self.feature_name = input("Name the constructed feature (column name), like 'NEW-COMPOUND': \n" "@input: ")
            if len(self.feature_name) == 0:
                print("Sorry!You haven't named it yet!")
            else:
                break
        return self.feature_name

    def input_expression(self) -> None:
        """Input the expression of the constructed feature."""
        expression = input(
            "Build up new feature with the combination of basic arithmatic operators,"
            " including '+', '-', '*', '/', '()'.\n"
            "Input example 1: a * b - c \n"
            "--> Step 1: Multiply a column with b column; \n"
            "--> Step 2: Subtract c from the result of Step 1; \n"
            "Input example 2: (d + 5 * f) / g \n"
            "--> Step 1: Multiply 5 with f; \n"
            "--> Step 2: Plus d column with the result of Step 1;\n"
            "--> Step 3: Divide the result of Step 1 by g; \n"
            "Input example 3: pow(a, b) + c * d \n"
            "--> Step 1: Raise the base a to the power of the exponent b; \n"
            "--> Step 2: Multiply the value of c by the value of d; \n"
            "--> Step 3: Add the result of Step 1 to the result of Step 2; \n"
            "Input example 4: log(a)/b - c \n"
            "--> Step 1: Take the logarithm of the value a; \n"
            "--> Step 2: Divide the result of Step 1 by the value of b; \n"
            "--> Step 3: Subtract the value of c from the result of Step 2; \n"
            "You can use mean(x) to calculate the average value.\n"
            "@input: "
        )
        while True:
            self._infix_expr = expression.replace(" ", "")
            if len(self._infix_expr) == 0:
                print("You haven't built any new features yet!")
                time.sleep(0.5)
                expression = input("-----* Please enter again *-----\n@input: ")
            elif not all(c.isdigit() or c.isspace() or c in FeatureConstructor.oper or c in self.alphabet for c in self._infix_expr):
                print("There's something wrong with the input !")
                time.sleep(0.5)
                expression = input("-----* Please enter again *-----\n@input: ")
            else:
                try:
                    self.letter_map()
                except Exception as e:
                    print("Your input contains the following error:", e)
                    time.sleep(0.5)
                    expression = input("-----* Please enter again *-----\n@input: ")
                else:
                    break
        return self._infix_expr

    def _evaluate_expression(self, feature_name, feature_expression) -> None:
        feature_expression = feature_expression.replace("sin", "np.sin")
        feature_expression = feature_expression.replace("cos", "np.cos")
        feature_expression = feature_expression.replace("tan", "np.tan")
        feature_expression = feature_expression.replace("pi", "np.pi")
        feature_expression = feature_expression.replace("pow", "np.power")
        feature_expression = feature_expression.replace("mean", "np.mean")
        feature_expression = feature_expression.replace("std", "np.std")
        feature_expression = feature_expression.replace("var", "np.var")
        feature_expression = feature_expression.replace("log", "np.log")
        try:
            new_feature_column = eval(feature_expression)
            if isinstance(new_feature_column, pd.Series) or isinstance(new_feature_column, pd.DataFrame):
                new_feature_column.name = feature_name
            else:
                new_feature_column = pd.Series([new_feature_column for _ in range(self.data.shape[0])], name=feature_name)
        except SyntaxError:
            print("The expression contains a syntax error.")
        except ZeroDivisionError:
            print("The expression contains a division by zero.")
        return new_feature_column

    def letter_map(self) -> None:
        """Map the letter to the column name."""
        new_text = ""
        test_text = "".join(ch for ch in self._infix_expr if ch not in set(" "))
        for words in FeatureConstructor.cal_words:
            if words in test_text:
                test_text = test_text.replace(words, ";" + words + ";")
        new_text = test_text
        self._infix_expr = ""
        if new_text[0] == ";":
            new_text = new_text[1:]
        for word in new_text.split(";"):
            if word in FeatureConstructor.cal_words:
                self._infix_expr += word
            else:
                for ww in word:
                    if ww in self.alphabet:
                        self._infix_expr += str("self.data['" + self._get_column(ww) + "']")
                    else:
                        self._infix_expr += ww

    def build(self) -> None:
        """Build the new feature."""
        print("The Selected Data Set:")
        show_data_columns(self.data.columns)
        fe_flag = 0
        is_feature_engineering = 0
        GEOPI_OUTPUT_ARTIFACTS_DATA_PATH = os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH")
        while True:
            if fe_flag != 1:
                print("Feature Engineering Option:")
                num2option(OPTION)
                is_feature_engineering = limit_num_input(OPTION, SECTION[1], num_input)
            if is_feature_engineering == 1:
                self.index2name()
                new_feature_name = self.input_feature_name()
                new_feature_expression = self.input_expression()
                instruction = {new_feature_name: new_feature_expression}
                self.config.update(instruction)
                new_feature_column = self._evaluate_expression(new_feature_name, new_feature_expression)
                clear_output()
                # Append the new feature to the original data and update the data
                self.data = self.append_feature(new_feature_column)
                clear_output()
                basic_info(self.data)
                basic_statistic(self.data)
                clear_output()
                print("Do you want to continue to build a new feature? ")
                num2option(OPTION)
                fe_flag = limit_num_input(OPTION, SECTION[1], num_input)
                if fe_flag == 1:
                    clear_output()
                    continue
                else:
                    save_data(self.data, "Data Selected Dropped-Imputed Feature-Engineering", GEOPI_OUTPUT_ARTIFACTS_DATA_PATH, MLFLOW_ARTIFACT_DATA_PATH)
                    print("Exit Feature Engineering Mode.")
                    clear_output()
                    break
            else:
                save_data(self.data, "Data Selected Dropped-Imputed Feature-Engineering", GEOPI_OUTPUT_ARTIFACTS_DATA_PATH, MLFLOW_ARTIFACT_DATA_PATH)
                clear_output()
                break
        return self.data

    def batch_build(self, feature_engineering_config: Dict) -> None:
        for feature_name, feature_expression in feature_engineering_config.items():
            new_feature_column = self._evaluate_expression(feature_name, feature_expression)
            self.data = self.append_feature(new_feature_column)
        return self.data

    def append_feature(self, new_feature_column: pd.Series) -> None:
        """Append the new feature to the original data."""
        print(f"Successfully construct a new feature {new_feature_column.name}.")
        print(new_feature_column)
        return pd.concat([self.data, new_feature_column], axis=1)



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from rich import print

from ..constants import MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH
from ..utils.base import save_data, save_fig


def basic_statistic(data: pd.DataFrame) -> None:
    """Some basic statistic information of the designated data set.

    Parameters
    ----------
    data : pd.DataFrame
        The data set.
    """
    print("Some basic statistic information of the designated data set:")
    print(data.describe())


def is_null_value(data: pd.DataFrame) -> None:
    """Check whether the data set has null value or not.

    Parameters
    ----------
    data : pd.DataFrame
        The data set.
    """
    print("Check which column has null values:")
    print("--" * 10)
    print(data.isnull().any())
    print("--" * 10)


def check_missing_value(data: pd.DataFrame) -> bool:
    """Check whether the data set has null value or not.

    Parameters
    ----------
    data : pd.DataFrame
        The data set.

    Returns
    -------
    flag : bool
        True if it has null value.
    """
    flag = data.isnull().any().any()
    if flag:
        print("Note: you'd better use imputation techniques to deal with the missing values.")
    else:
        print("Note: The provided data set is complete without missing values, we'll just pass this step!")
    return flag


def ratio_null_vs_filled(data: pd.DataFrame) -> None:
    """The ratio of the null values in each column.

    Parameters
    ----------
    data : pd.DataFrame
        The data set.
    """
    print("The ratio of the null values in each column:")
    print("--" * 10)
    print(data.isnull().mean().sort_values(ascending=False))
    print("--" * 10)


def correlation_plot(col: pd.Index, df: pd.DataFrame) -> None:
    """A heatmap describing the correlation between the required columns.

    Parameters
    ----------
    col : pd.Index
        A list of columns that need to plot.

    df : pd.DataFrame
        The data set.
    """
    plot_df = df[col]
    plot_df_cor = plot_df.corr()
    plt.figure(figsize=(20, 20))
    sns.heatmap(plot_df_cor, cmap="coolwarm", annot=True, linewidths=0.5)
    print("Successfully calculate the pair-wise correlation coefficient among the selected columns.")
    save_fig("Correlation Plot", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)
    save_data(df, "Correlation Plot", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)


def distribution_plot(col: pd.Index, df: pd.DataFrame) -> None:
    """The histogram containing the respective distribution subplots of the required columns.

    Parameters
    ----------
    col : pd.Index
        A list of columns that need to plot.

    df : pd.DataFrame
        The data set.
    """
    n = int(np.sqrt(len(col))) + 1
    plt.figure(figsize=(n * 2, n * 2))
    for i in range(len(col)):
        plt.subplot(n, n, i + 1)
        plt.hist(df[col[i]])
        plt.title(col[i])
    print("Successfully draw the distribution plot of the selected columns.")
    save_fig("Distribution Histogram", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)
    save_data(df, "Distribution Histogram", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)


def log_distribution_plot(col: pd.Index, df: pd.DataFrame) -> None:
    """The histogram containing the respective distribution subplots after log transformation of the required columns.

    Parameters
    ----------
    col : pd.Index
        A list of columns that need to plot.

    df : pd.DataFrame
        The data set.
    """
    # Log transform the required columns
    df_log_transformed = df[col].applymap(lambda x: np.log(x + 1))

    n = int(np.sqrt(len(col))) + 1
    plt.figure(figsize=(n * 2, n * 2))
    for i in range(len(col)):
        plt.subplot(n, n, i + 1)
        plt.hist(df_log_transformed[col[i]])
        plt.title(col[i])

    print("Successfully draw the distribution plot after log transformation of the selected columns.")
    save_fig("Distribution Histogram After Log Transformation", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)
    save_data(df_log_transformed, "Distribution Histogram After Log Transformation", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)


def probability_plot(col: pd.Index, df_origin: pd.DataFrame, df_impute: pd.DataFrame) -> None:
    """A large graph containing the respective probability plots (origin vs. impute) of the required columns.

    Parameters
    ----------
    col : pd.Index
        A list of columns that need to plot.

    df_origin : pd.DataFrame (n_samples, n_components)
        The original dataset with missing value.

    df_impute : pd.DataFrame (n_samples, n_components)
        The dataset after imputation.
    """
    r, c = len(col) // 4 + 1, 4
    fig = plt.figure(figsize=(c * 8, r * 8))
    for i in range(len(col)):
        feature = col[i]
        pp_origin = sm.ProbPlot(df_origin[feature].dropna(), fit=True)
        pp_impute = sm.ProbPlot(df_impute[feature], fit=True)
        ax = fig.add_subplot(r, c, i + 1)
        pp_origin.ppplot(line="45", other=pp_impute, ax=ax)
        plt.title(f"{feature}, origin data vs. imputed data")

    data = pd.concat([df_origin, df_impute], axis=1)
    print("Successfully draw the respective probability plot (origin vs. impute) of the selected columns")
    save_fig("Probability Plot", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)
    save_data(data, "Probability Plot", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_STATISTIC_PATH"), MLFLOW_ARTIFACT_IMAGE_STATISTIC_PATH)



from typing import Union

# from data_mining.schemas import Dataset
from pydantic import BaseModel


class UserBase(BaseModel):
    username: str = None
    email: str = None
    is_active: Union[bool, None] = True
    upload_count: int = 0


class UserCreate(UserBase):
    password: str


class User(UserBase):
    id: int
    hashed_password: str
    # datasets: list = [Dataset]

    class Config:
        orm_mode = True



# -*- coding: utf-8 -*-
import os

import pandas as pd
from multipledispatch import dispatch

from ..constants import MLFLOW_ARTIFACT_DATA_PATH
from ..model.classification import (
    ClassificationWorkflowBase,
    DecisionTreeClassification,
    ExtraTreesClassification,
    GradientBoostingClassification,
    KNNClassification,
    LogisticRegressionClassification,
    MLPClassification,
    RandomForestClassification,
    SGDClassification,
    SVMClassification,
    XGBoostClassification,
)
from ._base import ModelSelectionBase


class ClassificationModelSelection(ModelSelectionBase):
    """Simulate the normal way of training classification algorithms."""

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        self.clf_workflow = ClassificationWorkflowBase()
        self.transformer_config = {}

    @dispatch(object, object, object, object, object, object)
    def activate(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.DataFrame,
        y_test: pd.DataFrame,
    ) -> None:
        """Train by Scikit-learn framework."""

        self.clf_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        # Customize label
        y, y_train, y_test = self.clf_workflow.customize_label(y, y_train, y_test, os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH)

        # Sample balance
        sample_balance_config, X_train, y_train = self.clf_workflow.sample_balance(X_train, y_train, os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH)

        # Model option
        if self.model_name == "Support Vector Machine":
            hyper_parameters = SVMClassification.manual_hyper_parameters()
            self.clf_workflow = SVMClassification(
                kernel=hyper_parameters["kernel"],
                degree=hyper_parameters["degree"],
                gamma=hyper_parameters["gamma"],
                C=hyper_parameters["C"],
                shrinking=hyper_parameters["shrinking"],
            )
        elif self.model_name == "Decision Tree":
            hyper_parameters = DecisionTreeClassification.manual_hyper_parameters()
            self.clf_workflow = DecisionTreeClassification(
                criterion=hyper_parameters["criterion"],
                max_depth=hyper_parameters["max_depth"],
                min_samples_split=hyper_parameters["min_samples_split"],
                min_samples_leaf=hyper_parameters["min_samples_leaf"],
                max_features=hyper_parameters["max_features"],
            )
        elif self.model_name == "Random Forest":
            hyper_parameters = RandomForestClassification.manual_hyper_parameters()
            self.clf_workflow = RandomForestClassification(
                n_estimators=hyper_parameters["n_estimators"],
                max_depth=hyper_parameters["max_depth"],
                min_samples_split=hyper_parameters["min_samples_split"],
                min_samples_leaf=hyper_parameters["min_samples_leaf"],
                max_features=hyper_parameters["max_features"],
                bootstrap=hyper_parameters["bootstrap"],
                oob_score=hyper_parameters["oob_score"],
                max_samples=hyper_parameters["max_samples"],
            )
        elif self.model_name == "XGBoost":
            hyper_parameters = XGBoostClassification.manual_hyper_parameters()
            self.clf_workflow = XGBoostClassification(
                n_estimators=hyper_parameters["n_estimators"],
                learning_rate=hyper_parameters["learning_rate"],
                max_depth=hyper_parameters["max_depth"],
                subsample=hyper_parameters["subsample"],
                colsample_bytree=hyper_parameters["colsample_bytree"],
                alpha=hyper_parameters["alpha"],
                lambd=hyper_parameters["lambd"],
            )
        elif self.model_name == "Logistic Regression":
            hyper_parameters = LogisticRegressionClassification.manual_hyper_parameters()
            self.clf_workflow = LogisticRegressionClassification(
                penalty=hyper_parameters["penalty"],
                C=hyper_parameters["C"],
                solver=hyper_parameters["solver"],
                max_iter=hyper_parameters["max_iter"],
                class_weight=hyper_parameters["class_weight"],
                l1_ratio=hyper_parameters["l1_ratio"],
            )
        elif self.model_name == "Multi-layer Perceptron":
            hyper_parameters = MLPClassification.manual_hyper_parameters()
            self.clf_workflow = MLPClassification(
                hidden_layer_sizes=hyper_parameters["hidden_layer_sizes"],
                activation=hyper_parameters["activation"],
                solver=hyper_parameters["solver"],
                alpha=hyper_parameters["alpha"],
                learning_rate=hyper_parameters["learning_rate"],
                max_iter=hyper_parameters["max_iter"],
            )
        elif self.model_name == "Extra-Trees":
            hyper_parameters = ExtraTreesClassification.manual_hyper_parameters()
            self.clf_workflow = ExtraTreesClassification(
                n_estimators=hyper_parameters["n_estimators"],
                max_depth=hyper_parameters["max_depth"],
                min_samples_split=hyper_parameters["min_samples_split"],
                min_samples_leaf=hyper_parameters["min_samples_leaf"],
                max_features=hyper_parameters["max_features"],
                bootstrap=hyper_parameters["bootstrap"],
                oob_score=hyper_parameters["oob_score"],
                max_samples=hyper_parameters["max_samples"],
            )
        elif self.model_name == "Gradient Boosting":
            hyper_parameters = GradientBoostingClassification.manual_hyper_parameters()
            self.clf_workflow = GradientBoostingClassification(
                n_estimators=hyper_parameters["n_estimators"],
                learning_rate=hyper_parameters["learning_rate"],
                max_depth=hyper_parameters["max_depth"],
                min_samples_split=hyper_parameters["min_samples_split"],
                min_samples_leaf=hyper_parameters["min_samples_leaf"],
                max_features=hyper_parameters["max_features"],
                subsample=hyper_parameters["subsample"],
                loss=hyper_parameters["loss"],
            )
        elif self.model_name == "K-Nearest Neighbors":
            hyper_parameters = KNNClassification.manual_hyper_parameters()
            self.clf_workflow = KNNClassification(
                n_neighbors=hyper_parameters["n_neighbors"],
                weights=hyper_parameters["weights"],
                algorithm=hyper_parameters["algorithm"],
                leaf_size=hyper_parameters["leaf_size"],
                p=hyper_parameters["p"],
                metric=hyper_parameters["metric"],
            )
        elif self.model_name == "Stochastic Gradient Descent":
            hyper_parameters = SGDClassification.manual_hyper_parameters()
            self.clf_workflow = SGDClassification(
                loss=hyper_parameters["loss"],
                penalty=hyper_parameters["penalty"],
                alpha=hyper_parameters["alpha"],
                l1_ratio=hyper_parameters["l1_ratio"],
                fit_intercept=hyper_parameters["fit_intercept"],
                max_iter=hyper_parameters["max_iter"],
                tol=hyper_parameters["tol"],
                shuffle=hyper_parameters["shuffle"],
                learning_rate=hyper_parameters["learning_rate"],
                eta0=hyper_parameters["eta0"],
                power_t=hyper_parameters["power_t"],
                early_stopping=hyper_parameters["early_stopping"],
                validation_fraction=hyper_parameters["validation_fraction"],
                n_iter_no_change=hyper_parameters["n_iter_no_change"],
            )
        self.clf_workflow.show_info()

        # Use Scikit-learn style API to process input data
        self.clf_workflow.fit(X_train, y_train)
        y_train_predict = self.clf_workflow.predict(X_train)
        y_train_predict = self.clf_workflow.np2pd(y_train_predict, y_train.columns)
        self.clf_workflow.data_upload(y_train_predict=y_train_predict)
        y_test_predict = self.clf_workflow.predict(X_test)
        y_test_predict = self.clf_workflow.np2pd(y_test_predict, y_test.columns)
        self.clf_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test, y_test_predict=y_test_predict)

        # Save the model hyper-parameters
        self.clf_workflow.save_hyper_parameters(hyper_parameters, self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))

        # Common components for every classification algorithm
        self.clf_workflow.common_components()

        # Special components of different algorithms
        self.clf_workflow.special_components()

        # Save the prediction result
        self.clf_workflow.data_save(y_train_predict, "Y Train Predict", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Model Train Prediction")
        self.clf_workflow.data_save(y_test_predict, "Y Test Predict", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Model Test Prediction")

        # Save the trained model
        self.clf_workflow.model_save()

    @dispatch(object, object, object, object, object, object, bool)
    def activate(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        X_train: pd.DataFrame,
        X_test: pd.DataFrame,
        y_train: pd.DataFrame,
        y_test: pd.DataFrame,
        is_automl: bool,
    ) -> None:
        """Train by FLAML framework + RAY framework."""

        self.clf_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test)

        # Customize label
        y, y_train, y_test = self.clf_workflow.customize_label(y, y_train, y_test, os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH)

        # Sample balance
        sample_balance_config, X_train, y_train = self.clf_workflow.sample_balance(X_train, y_train, os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH)

        # Model option
        if self.model_name == "Support Vector Machine":
            self.clf_workflow = SVMClassification()
        elif self.model_name == "Decision Tree":
            self.clf_workflow = DecisionTreeClassification()
        elif self.model_name == "Random Forest":
            self.clf_workflow = RandomForestClassification()
        elif self.model_name == "XGBoost":
            self.clf_workflow = XGBoostClassification()
        elif self.model_name == "Logistic Regression":
            self.clf_workflow = LogisticRegressionClassification()
        elif self.model_name == "Multi-layer Perceptron":
            self.clf_workflow = MLPClassification()
        elif self.model_name == "Extra-Trees":
            self.clf_workflow = ExtraTreesClassification()
        elif self.model_name == "Gradient Boosting":
            self.clf_workflow = GradientBoostingClassification()
        elif self.model_name == "K-Nearest Neighbors":
            self.clf_workflow = KNNClassification()
        elif self.model_name == "Stochastic Gradient Descent":
            self.clf_workflow = SGDClassification()

        self.clf_workflow.show_info()

        # Use Scikit-learn style API to process input data
        self.clf_workflow.fit(X_train, y_train, is_automl)
        y_train_predict = self.clf_workflow.predict(X_train, is_automl)
        y_train_predict = self.clf_workflow.np2pd(y_train_predict, y_train.columns)
        self.clf_workflow.data_upload(y_train_predict=y_train_predict)
        y_test_predict = self.clf_workflow.predict(X_test, is_automl)
        y_test_predict = self.clf_workflow.np2pd(y_test_predict, y_test.columns)
        self.clf_workflow.data_upload(X=X, y=y, X_train=X_train, X_test=X_test, y_train=y_train, y_test=y_test, y_test_predict=y_test_predict)

        # Save the model hyper-parameters
        if self.clf_workflow.ray_best_model is not None:
            self.clf_workflow.save_hyper_parameters(self.clf_workflow.ray_best_model.get_params(), self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))
        else:
            self.clf_workflow.save_hyper_parameters(self.clf_workflow.automl.best_config, self.model_name, os.getenv("GEOPI_OUTPUT_PARAMETERS_PATH"))

        # Common components for every classification algorithm
        self.clf_workflow.common_components(is_automl)

        # Special components of different algorithms
        self.clf_workflow.special_components(is_automl)

        # Save the prediction result
        self.clf_workflow.data_save(y_train_predict, "Y Train Predict", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Model Train Prediction")
        self.clf_workflow.data_save(y_test_predict, "Y Test Predict", os.getenv("GEOPI_OUTPUT_ARTIFACTS_DATA_PATH"), MLFLOW_ARTIFACT_DATA_PATH, "Model Test Prediction")

        # Save the trained model
        self.clf_workflow.model_save(is_automl)



# -*- coding: utf-8 -*-
from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input


def tsne_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Components: This parameter specifies the number of components to retain after dimensionality reduction.")
    print("Please specify the number of components to retain. A good starting range could be between 2 and 10, such as 4.")
    n_components = num_input(SECTION[2], "N Components: ")
    print("Perplexity: This parameter is related to the number of nearest neighbors that each point considers when computing the probabilities.")
    print("Please specify the perplexity. A good starting range could be between 5 and 50, such as 30.")
    perplexity = num_input(SECTION[2], "Perplexity: ")
    print("Learning Rate: This parameter controls the step size during the optimization process.")
    print("Please specify the learning rate. A good starting range could be between 10 and 1000, such as 200.")
    learning_rate = float_input(200, SECTION[2], "Learning Rate: ")
    print("Number of Iterations: This parameter controls how many iterations the optimization will run for.")
    print("Please specify the number of iterations. A good starting range could be between 250 and 1000, such as 500. The minimum is 250.")
    is_valid = False
    while not is_valid:
        n_iter = num_input(SECTION[2], "Number of Iterations: ")
        if n_iter >= 250:
            is_valid = True
        else:
            print("Please enter a number greater than or equal to 250.")
    print("Early Exaggeration: This parameter controls how tight natural clusters in the original space are in the embedded space and how much space will be between them.")
    print("Please specify the early exaggeration. A good starting range could be between 5 and 50, such as 12.")
    early_exaggeration = float_input(12, SECTION[2], "Early Exaggeration: ")
    hyper_parameters = {"n_components": n_components, "perplexity": perplexity, "learning_rate": learning_rate, "n_iter": n_iter, "early_exaggeration": early_exaggeration}
    return hyper_parameters



from datetime import timedelta

from database import get_db
from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import OAuth2PasswordRequestForm
from sqlalchemy.orm import Session

from .constants import ACCESS_TOKEN_EXPIRE_MINUTES
from .dependencies import get_current_active_user
from .schemas import User, UserCreate
from .service import authenticate_user, create_new_user, get_user_by_email, get_user_by_username
from .utils import create_access_token

router = APIRouter(
    prefix="/auth",
    tags=["auth"],
    responses={404: {"description": "Not found"}},
)


# @router.get("/", response_model=List[User])
# async def read_users(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
#     users = get_users(db, skip=0, limit=100)
#     return users


@router.get("/me")
async def read_users_me(current_user: User = Depends(get_current_active_user)):
    user_info = {"username": current_user.username, "email": current_user.email, "upload_count": current_user.upload_count}
    return user_info


# @router.get("/{user_id}", response_model=User)
# async def read_user(user_id: int, db: Session = Depends(get_db)):
#     db_user = get_user_by_id(db, user_id)
#     if db_user is None:
#         raise HTTPException(status_code=404, detail="User not found")
#     return db_user


@router.put("/{username}")
async def update_user(username: str):
    return {"error": "Not implemented"}


@router.post("/login")
async def login_for_access_token(form_data: OAuth2PasswordRequestForm = Depends(), db: Session = Depends(get_db)):
    # OAuth2PasswordRequestForm is a class that has username and password attributes
    # The user is only allowed to login with email, so we need to change the username to email
    user_email = form_data.username
    user = authenticate_user(db, user_email, form_data.password)
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        # data={"sub": user.username}, expires_delta=access_token_expires
        data={"sub": user.email},
        expires_delta=access_token_expires,
    )
    return {"access_token": access_token, "token_type": "bearer"}


@router.post("/register")
async def register(email: str, form_data: OAuth2PasswordRequestForm = Depends(), db: Session = Depends(get_db)):

    db_user = get_user_by_username(db, username=form_data.username)
    if db_user:
        raise HTTPException(status_code=400, detail="Username already taken")
    db_user = get_user_by_email(db, email=email)
    if db_user:
        raise HTTPException(status_code=400, detail="Email already registered")

    user = UserCreate(email=email, username=form_data.username, password=form_data.password)
    create_new_user(db=db, user=user)
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        # data={"sub": user.username}, expires_delta=access_token_expires
        data={"sub": user.email},
        expires_delta=access_token_expires,
    )
    return {"access_token": access_token, "token_type": "bearer"}



from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rich import print

from ....constants import SECTION
from ....data.data_readiness import float_input, num_input, str_input


def logistic_regression_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("Penalty: This hyperparameter specifies the norm used in the penalization.")
    print("Please specify the norm used in the penalization. It is generally recommended to leave it as 'l2'.")
    penalties = ["l1", "l2", "elasticnet", "None"]
    penalty = str_input(penalties, SECTION[2])
    if penalty == "None":
        penalty = None
    print("C: This hyperparameter specifies the inverse of regularization strength. A smaller value of C indicates stronger regularization, whereas a larger value indicates weaker regularization.")
    print("Please specify the inverse of regularization strength. A good starting range could be between 0.001 and 1000, such as 1.0.")
    C = float_input(1, SECTION[2], "@C: ")
    l1_ratio = None
    if penalty == "l1":
        print("Solver: This hyperparameter specifies the algorithm to use in the optimization problem.")
        print("Please specify the algorithm to use in the optimization problem. It is generally recommended to leave it as 'liblinear'.")
        solvers = ["liblinear", "saga"]
        solver = str_input(solvers, SECTION[2])
    elif penalty == "l2" or penalty == "none":
        print("Solver: This hyperparameter specifies the algorithm to use in the optimization problem.")
        print("Please specify the algorithm to use in the optimization problem. It is generally recommended to leave it as 'lbfgs'.")
        solvers = ["newton-cg", "lbfgs", "sag", "saga"]
        solver = str_input(solvers, SECTION[2])
    elif penalty == "elasticnet":
        solver = "saga"
        print("L1 Ratio: This hyperparameter specifies the Elastic-Net mixing parameter, with 0 <= l1_ratio <= 1.")
        print("Please specify the Elastic-Net mixing parameter. A good starting range could be between 0.0 and 1.0, such as 0.5.")
        l1_ratio = float_input(0.5, SECTION[2], "@L1 Ratio: ")
    print("Max Iter: This hyperparameter specifies the maximum number of iterations taken for the solvers to converge.")
    print("Please specify the maximum number of iterations taken for the solvers to converge. A good starting range could be between 100 and 1000, such as 100.")
    max_iter = num_input(SECTION[2], "@Max Iter: ")
    print(
        "Class Weight: This hyperparameter specifies the weights associated with classes. It can be set to 'balanced'"
        " to automatically adjust the weights inversely proportional to the class frequencies in the input data."
    )
    print("Please specify the weights associated with classes. It is generally recommended to leave it as None.")
    class_weights = ["None", "balanced"]
    class_weight = str_input(class_weights, SECTION[2])
    if class_weight == "None":
        class_weight = None
    hyper_parameters = {
        "penalty": penalty,
        "C": C,
        "solver": solver,
        "max_iter": max_iter,
        "class_weight": class_weight,
        "l1_ratio": l1_ratio,
    }
    return hyper_parameters


def plot_logistic_importance(columns_name: np.ndarray, trained_model: object) -> pd.DataFrame:
    """Draw the feature importance diagram for analysis.

    Parameters
    ----------
    data: pd.DataFrame (n_samples, n_components)
        Data for silhouette.

    trained_model: any
        The algorithm which to be used.

    References
    ----------
    Logistic regression, despite its name, is a linear model for classification rather than regression.
    Logistic regression is also known in the literature as logit regression, maximum-entropy classific
    ation (MaxEnt) or the log-linear classifier. In this model, the probabilities describing the possible
    outcomes of a single trial are modeled using a logistic function.

    https://scikit-learn.org/stable/modules/linear_model.html/logistic-regression
    """
    for feature_name, score in zip(list(columns_name), trained_model.coef_.flatten()):
        print(feature_name, ":", score)

    # feature importance map ranked by coefficient
    coef_lr = pd.DataFrame({"var": columns_name, "coef": trained_model.coef_.flatten()})
    index_sort = np.abs(coef_lr["coef"]).sort_values().index
    coef_lr_sort = coef_lr.loc[index_sort, :]

    # Horizontal column chart plot
    fig, ax = plt.subplots(figsize=(14, 8))
    x, y = coef_lr_sort["var"], coef_lr_sort["coef"]
    rects = plt.barh(x, y, color="dodgerblue")
    plt.grid(linestyle="-.", axis="y", alpha=0.4)
    plt.tight_layout()

    # Add data labels
    for rect in rects:
        w = rect.get_width()
        ax.text(w, rect.get_y() + rect.get_height() / 2, "%.2f" % w, ha="left", va="center")
        plt.title("Feature Importance Map Ranked by Coefficient")

    return coef_lr_sort



# -*- coding: utf-8 -*-
import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rich import print

from ..constants import MLFLOW_ARTIFACT_IMAGE_MAP_PATH, OPTION, SECTION
from ..data.data_readiness import limit_num_input, num2option, num_input, show_data_columns
from ..utils.base import clear_output, get_os, save_data, save_fig

logging.captureWarnings(True)


def map_projected_by_cartopy(col: pd.Series, longitude: pd.DataFrame, latitude: pd.DataFrame) -> None:
    """Project an element data into world map using cartopy.

    Parameters
    ----------
    col : pd.Series
        One selected column from the data sheet.

    longitude : pd.DataFrame
        Longitude data of data items.

    latitude : pd.DataFrame
        Latitude data of data items.
    """
    try:
        import cartopy
        import cartopy.crs as ccrs
    except ImportError:
        print("The cartopy package is not installed. Please install it first.")
        print("For example, you can run the following command in your terminal:")
        print("pip install cartopy")
        print("conda install -c conda-forge cartopy")
        return
    M = col
    # Create a new figure with the desired size and DPI
    plt.figure(figsize=(24, 16), dpi=300)
    # Set the font style
    plt.rcParams["font.sans-serif"] = "Arial"

    # Create a Robinson projection centered at the equator
    projection = ccrs.Robinson(central_longitude=0, globe=None, false_easting=0, false_northing=0)
    # Create a new axis with the Robinson projection
    ax = plt.axes(projection=projection)

    # Add coastlines and borders
    ax.coastlines()
    ax.add_feature(cartopy.feature.BORDERS)

    # Set the map boundaries and fill color
    ax.set_global()
    ax.set_facecolor("white")

    # Define parallels and meridians
    parallels = np.arange(-90.0, 90.0, 45.0)
    meridians = np.arange(-180.0, 180.0, 60.0)

    # Draw parallels and meridians
    ax.gridlines(draw_labels=True, linewidth=0.5, color="gray", alpha=0.5, linestyle="--", xlocs=meridians, ylocs=parallels)

    # Set the color values
    if type(M) != type(longitude):
        M = [0.5 for i in range(len(M))]

    # Create a scatter plot with color and size parameters
    sc = ax.scatter(longitude, latitude, c=pd.DataFrame(M), edgecolor="grey", transform=ccrs.PlateCarree(), linewidths=0.5, vmax=3, vmin=0, s=25, alpha=0.6, cmap="BuPu")

    # Add colorbar
    cbar = plt.colorbar(sc, ax=ax, fraction=0.01, pad=0.1)
    cbar.ax.tick_params(labelsize=30)
    cbar.set_label("Counts", fontsize=30)

    # save figure and data
    data = pd.concat([col, longitude, latitude], axis=1)
    save_fig(f"Map Projection - {col.name}", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MAP_PATH"), MLFLOW_ARTIFACT_IMAGE_MAP_PATH)
    save_data(data, f"Map Projection - {col.name}", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MAP_PATH"), MLFLOW_ARTIFACT_IMAGE_MAP_PATH)


def map_projected_by_basemap(col: pd.Series, longitude: pd.DataFrame, latitude: pd.DataFrame) -> None:
    """Project an element data into world map using basemap.

    Parameters
    ----------
    col : pd.Series
        One selected column from the data sheet.

    longitude : pd.DataFrame
        Longitude data of data items.

    latitude : pd.DataFrame
        Latitude data of data items.
    """
    try:
        from mpl_toolkits.basemap import Basemap
    except ImportError:
        print("The basemap package is not installed. Please install it first.")
        print("For example, you can run the following command in your terminal:")
        print("pip install basemap")
        print("conda install -c conda-forge basemap")
        return
    M = col
    plt.figure(figsize=(24, 16), dpi=300)
    plt.rcParams["font.sans-serif"] = "Arial"
    m = Basemap(projection="robin", lat_0=0, lon_0=0)
    m.drawcoastlines()
    m.drawcountries()
    m.drawmapboundary(fill_color="white")

    parallels = np.arange(-90.0, 90.0, 45.0)
    m.drawparallels(parallels, labels=[True, True, True, False], fontsize=30)
    meridians = np.arange(-180.0, 180.0, 60.0)
    m.drawmeridians(meridians, labels=[True, False, True, True], fontsize=30)
    lon, lat = m(longitude, latitude)
    if type(M) != type(longitude):
        M = [0.5 for i in range(len(M))]
    m.scatter(lon, lat, c=M, edgecolor="grey", marker="D", linewidths=0.5, vmax=3, vmin=0, s=25, alpha=0.3, cmap="BuPu")
    cb = m.colorbar(pad=1)
    cb.ax.tick_params(labelsize=30)
    cb.set_label("Counts", fontsize=50)

    data = pd.concat([col, longitude, latitude], axis=1)
    save_fig(f"Map Projection - {col.name}", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MAP_PATH"), MLFLOW_ARTIFACT_IMAGE_MAP_PATH)
    save_data(data, f"Map Projection - {col.name}", os.getenv("GEOPI_OUTPUT_ARTIFACTS_IMAGE_MAP_PATH"), MLFLOW_ARTIFACT_IMAGE_MAP_PATH)


def process_world_map(data: pd.DataFrame) -> None:
    """The process of projecting the data on the world map."""
    map_flag = 0
    is_map_projection = 0
    detection_index = 0
    lon = ["LONGITUDE", "Longitude (°E)", "longitude", "Longitude", "经度 (°N)", "经度", "lng"]
    lat = ["LATITUDE", "Latitude (°N)", "latitude", "Latitude", "纬度 (°E)", "纬度", "lat"]
    j = [j for j in lat if j in data.columns]
    i = [i for i in lon if i in data.columns]
    if bool(len(j) > 0):
        detection_index += 1
    if bool(len(i) > 0):
        detection_index += 2
    if detection_index == 2:
        print("The provided data set is lack of 'LATITUDE' data.")
    elif detection_index == 1:
        print("The provided data set is lack of 'LONGITUDE' data.")
    elif detection_index == 0:
        print("The provided data set is lack of 'LONGITUDE' and 'LATITUDE' data.")
    if detection_index != 3:
        print("Hence, world map projection functionality will be skipped!")
        clear_output()
    # If the data set contains both longitude and latitude data, then the user can choose to project the data on the world map.
    while detection_index == 3:
        if map_flag != 1:
            # Check if the user wants to project the data on the world map.
            print("World Map Projection for A Specific Element Option:")
            num2option(OPTION)
            is_map_projection = limit_num_input(OPTION, SECTION[3], num_input)
            clear_output()
        if is_map_projection == 1:
            # If the user chooses to project the data on the world map, then the user can select the element to be projected.
            print("-*-*- Distribution in World Map -*-*-")
            print("Select one of the elements below to be projected in the World Map: ")
            show_data_columns(data.columns)
            elm_num = limit_num_input(data.columns, SECTION[3], num_input)
            clear_output()
            latitude = data.loc[:, j]
            longitude = data.loc[:, i]
            print("Longitude and latitude data are selected from the provided data set.")
            # If OS is Windows or Linux, then use basemap to project the data on the world map.
            # If OS is macOS, then use cartopy to project the data on the world map.
            my_os = get_os()
            if my_os == "Windows" or my_os == "Linux":
                map_projected_by_basemap(data.iloc[:, elm_num - 1], longitude, latitude)
            elif my_os == "macOS":
                map_projected_by_cartopy(data.iloc[:, elm_num - 1], longitude, latitude)
            clear_output()
            print("Do you want to continue to project a new element in the World Map?")
            num2option(OPTION)
            map_flag = limit_num_input(OPTION, SECTION[3], num_input)
            if map_flag == 1:
                clear_output()
                continue
            else:
                print("Exit Map Projection Mode.")
                clear_output()
                break
        elif is_map_projection == 2:
            break


# def map_projected(col: pd.Series, longitude: pd.DataFrame, latitude: pd.DataFrame) -> None:
#     """Project an element data into world map.

#     Parameters
#     ----------
#     col : pd.Series
#         One selected column from the data sheet.

#     longitude : pd.DataFrame
#         Longitude data of data items.

#     latitude : pd.DataFrame
#         Latitude data of data items.
#     """
#     # Create point geometries
#     geometry = geopandas.points_from_xy(longitude, latitude)

#     geo_df = geopandas.GeoDataFrame(pd.concat([col, longitude, latitude], axis=1), geometry=geometry)
#     world = geopandas.read_file(geopandas.datasets.get_path('naturalearth_lowres'))

#     # Make figure
#     fig, ax = plt.subplots(figsize=(24, 18))
#     divider = make_axes_locatable(ax)
#     cax = divider.append_axes("right", size="5%", pad=0.1)
#     world.plot(ax=ax, alpha=0.4, color='grey', edgecolor='black')
#     geo_df.plot(col.name, ax=ax, s=20, cmap='gist_heat_r', cax=cax, legend=True)
#     plt.title('colorbar')
#     save_fig(f"Map Projection - {col.name}", MAP_IMAGE_PATH)



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input


def bayesian_ridge_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters for Bayesian Ridge Regression.

    Returns
    -------
    hyper_parameters : dict
        The hyperparameters.
    """
    # print("Maximum Number of Iterations: The maximum number of iterations for training. The default is 300.")
    # max_iter = num_input(SECTION[2], "@Max Iterations: ")

    print("Tolerance: The tolerance for stopping criteria. A good starting value could be between 1e-6 and 1e-3, such as 1e-4. The default is 0.001.")
    tol = float_input(1e-4, SECTION[2], "@Tol: ")

    print("Alpha 1: Hyperparameter for the Gamma distribution prior over the alpha parameter. A good starting value could be between 1e-6 and 1e-3, such as 1e-6. The default is 0.000001.")
    alpha_1 = float_input(1e-6, SECTION[2], "@Alpha 1: ")

    print("Alpha 2: Hyperparameter for the Gamma distribution prior over the beta parameter. A good starting value could be between 1e-6 and 1e-3, such as 1e-6. The default is 0.000001.")
    alpha_2 = float_input(1e-6, SECTION[2], "@Alpha 2: ")

    print("Lambda 1: Hyperparameter for the Gaussian distribution prior over the weights. A good starting value could be between 1e-6 and 1e-3, such as 1e-6. The default is 0.000001.")
    lambda_1 = float_input(1e-6, SECTION[2], "@Lambda 1: ")

    print("Lambda 2: Hyperparameter for the Gamma distribution prior over the noise. A good starting value could be between 1e-6 and 1e-3, such as 1e-6. The default is 0.000001.")
    lambda_2 = float_input(1e-6, SECTION[2], "@Lambda 2: ")

    print("Alpha Init: Initial guess for alpha. The default is 0.000001.")
    alpha_init = float_input(1.0, SECTION[2], "@Alpha Init: ")

    print("Lambda Init: Initial guess for lambda. The default is 0.000001.")
    lambda_init = float_input(1.0, SECTION[2], "@Lambda Init: ")

    print("Compute Score: Whether to compute the objective function at each step. The default is False.")
    compute_score = bool_input(SECTION[2])

    print("Fit Intercept: Whether to fit an intercept term. It is generally recommended to set it as True.")
    fit_intercept = bool_input(SECTION[2])

    print("Copy X: Whether to copy X in fit. The default is True.")
    copy_X = bool_input(SECTION[2])

    print("Verbose: Verbosity of the solution. The default is False.")
    verbose = bool_input(SECTION[2])

    hyper_parameters = {
        "tol": tol,
        "alpha_1": alpha_1,
        "alpha_2": alpha_2,
        "lambda_1": lambda_1,
        "lambda_2": lambda_2,
        "alpha_init": alpha_init,
        "lambda_init": lambda_init,
        "compute_score": compute_score,
        "fit_intercept": fit_intercept,
        "copy_X": copy_X,
        "verbose": verbose,
    }
    return hyper_parameters



from datetime import datetime, timedelta
from typing import Union

from fastapi import HTTPException, status
from jose import JWTError, jwt
from passlib.context import CryptContext

from .constants import ALGORITHM, SECRET_KEY

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def get_password_hash(password):
    return pwd_context.hash(password)


def verify_password(plain_password, hashed_password):
    return pwd_context.verify(plain_password, hashed_password)


def create_access_token(data: dict, expires_delta: Union[timedelta, None] = None):
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


def decode_access_token(token: str):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        # username: str = payload.get("sub")
        email: str = payload.get("sub")
        # if username is None:
        if email is None:
            raise credentials_exception
        # token_data = TokenData(username=username)
        # token_data = TokenData(email=email)
        token_data = email
    except JWTError:
        raise credentials_exception
    return token_data



from typing import Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rich import print
from sklearn.inspection import permutation_importance
from sklearn.tree import plot_tree

# <------
# Used by tree-based models including classification and regression besides XGBoost


def plot_decision_tree(trained_model: object, image_config: Dict) -> None:
    """Drawing decision tree diagrams.

    Parameters
    ----------
    trained_model : sklearn algorithm model
        The sklearn algorithm model trained with X_train data.

    image_config : dict
        Image Configuration
    """
    # create drawing canvas
    fig, ax = plt.subplots(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])

    # draw the main content
    plot_tree(
        trained_model,
        max_depth=image_config["max_depth"],
        feature_names=image_config["feature_names"],
        class_names=image_config["class_names"],
        label=image_config["label"],
        filled=image_config["filled"],
        impurity=image_config["impurity"],
        node_ids=image_config["node_ids"],
        proportion=image_config["proportion"],
        rounded=image_config["rounded"],
        precision=image_config["precision"],
        ax=image_config["ax"],
        fontsize=image_config["fontsize"],
    )

    # automatically optimize picture layout structure
    fig.tight_layout()
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    x_adjustment = (xmax - xmin) * 0.01
    y_adjustment = (ymax - ymin) * 0.01
    ax.axis([xmin - x_adjustment, xmax + x_adjustment, ymin - y_adjustment, ymax + y_adjustment])

    # convert the font of the axes
    # plt.tick_params(labelsize=image_config['labelsize'])  # adjust the font size of the axis label
    # plt.setp(ax.get_xticklabels(), rotation=image_config['xrotation'], ha=image_config['xha'],
    #          rotation_mode="anchor")  # axis label rotation Angle
    # plt.setp(ax.get_yticklabels(), rotation=image_config['rot'], ha=image_config['yha'],
    #          rotation_mode="anchor")  # axis label rotation Angle
    x1_label = ax.get_xticklabels()  # adjust the axis label font
    [x1_label_temp.set_fontname(image_config["axislabelfont"]) for x1_label_temp in x1_label]
    y1_label = ax.get_yticklabels()
    [y1_label_temp.set_fontname(image_config["axislabelfont"]) for y1_label_temp in y1_label]

    ax.set_title(
        label=image_config["title_label"],
        fontdict={
            "size": image_config["title_size"],
            "color": image_config["title_color"],
            "family": image_config["title_font"],
        },
        loc=image_config["title_location"],
        pad=image_config["title_pad"],
    )


# Used by tree-based models including classification and regression besides XGBoost
# ------>

# <------
# Used by tree-based models, like, random forest, extra-trees, xgboost including classification and regression


def plot_feature_importance(columns_name: pd.Index, feature_importance: np.ndarray, image_config: dict) -> pd.DataFrame:
    """Draw the feature importance bar diagram.

    Parameters
    ----------
    columns_name : pd.Index
        The name of the columns.

    feature_importance : np.ndarray
        The feature importance values.

    image_config : dict
        The configuration of the image.

    Returns
    -------
    importance : pd.DataFrame
        The feature importance values.
    """
    # create drawing canvas
    fig, ax = plt.subplots(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])

    # print the feature importance value orderly
    for feature_name, score in zip(list(columns_name), feature_importance):
        print(feature_name, ":", score)

    # draw the main content
    importance = pd.DataFrame({"Feature": columns_name, "Importance": feature_importance})
    importance = importance.sort_values(["Importance"], ascending=True)
    importance["Importance"] = (importance["Importance"]).astype(float)
    importance = importance.sort_values(["Importance"])
    importance.set_index("Feature", inplace=True)
    importance.plot.barh(alpha=image_config["alpha2"], rot=0)

    # automatically optimize picture layout structure
    fig.tight_layout()
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    x_adjustment = (xmax - xmin) * 0.01
    y_adjustment = (ymax - ymin) * 0.01
    ax.axis([xmin - x_adjustment, xmax + x_adjustment, ymin - y_adjustment, ymax + y_adjustment])

    # convert the font of the axes
    x1_label = ax.get_xticklabels()  # adjust the axis label font
    [x1_label_temp.set_fontname(image_config["axislabelfont"]) for x1_label_temp in x1_label]
    y1_label = ax.get_yticklabels()
    [y1_label_temp.set_fontname(image_config["axislabelfont"]) for y1_label_temp in y1_label]

    ax.set_title(
        label=image_config["title_label"],
        fontdict={
            "size": image_config["title_size"],
            "color": image_config["title_color"],
            "family": image_config["title_font"],
        },
        loc=image_config["title_location"],
        pad=image_config["title_pad"],
    )

    return importance


# Used by tree-based models, like, random forest, extra-trees, xgboost including classification and regression
# ------>

# <------
# Used by linear models including classification and regression


def show_formula(coef: np.ndarray, intercept: np.ndarray, features_name: np.ndarray, regression_classification: str, y_train: pd.DataFrame) -> Dict:
    """Show the formula of linear models.

    Parameters
    ----------
    coef : np.ndarray
        Coefficient of the features in the decision function.

    intercept : np.ndarray or float
        Independent term in decision function.

    features_name : np.ndarray
        Name of the features.

    regression_classification : str
        Indicates whether it's a regression or classification model.

    y_train : pd.DataFrame
        The train label data.

    Returns
    -------
    formula : dict
        The formula of linear models.
    """
    formula = {}

    if regression_classification == "Regression":
        if len(y_train.columns) == 1:  # Single target
            coef = np.around(coef, decimals=3)[0]
            # Check if intercept is a scalar
            if isinstance(intercept, np.ndarray):
                intercept = np.around(intercept, decimals=3)[0]
            else:
                intercept = np.around(intercept, decimals=3)

            terms = [("-" if c < 0 else "+") + " " + str(abs(c)) + f if c != 0 else "" for c, f in zip(coef, features_name)]
            terms_first = (terms[0][2:] if coef[0] > 0 else terms[0]).replace(" ", "")
            formula["y:"] = terms_first + " " + " ".join(terms[1:]) + (" - " if intercept < 0 else " + ") + str(abs(intercept))
            print("y = ", formula["y:"])

        else:  # Multiple targets
            coef = np.around(coef, decimals=3)
            intercept = np.around(intercept, decimals=3)

            for idx, (coef_temp, intercept_temp) in enumerate(zip(coef, intercept)):
                terms_temp = [("-" if c < 0 else "+") + " " + str(abs(c)) + f if c != 0 else "" for c, f in zip(coef_temp, features_name)]
                terms_temp_first = (terms_temp[0][2:] if coef_temp[0] > 0 else terms_temp[0]).replace(" ", "")
                formula["y (" + y_train.columns[idx] + ") = "] = terms_temp_first + " " + " ".join(terms_temp[1:]) + (" - " if intercept_temp < 0 else " + ") + str(abs(intercept_temp))
                print("y (" + y_train.columns[idx] + ") = ", formula["y (" + y_train.columns[idx] + ") = "])

    elif regression_classification == "Classification":
        if coef.shape[0] == 1:  # Binary classification
            coef = np.around(coef, decimals=3)[0]
            # Check if intercept is a scalar
            if isinstance(intercept, np.ndarray):
                intercept = np.around(intercept, decimals=3)[0]
            else:
                intercept = np.around(intercept, decimals=3)

            terms = [("-" if c < 0 else "+") + " " + str(abs(c)) + f if c != 0 else "" for c, f in zip(coef, features_name)]
            terms_first = (terms[0][2:] if coef[0] > 0 else terms[0]).replace(" ", "")
            formula["y:"] = terms_first + " " + " ".join(terms[1:]) + (" - " if intercept < 0 else " + ") + str(abs(intercept))
            print("y = ", formula["y:"])

        else:  # Multiclass classification
            label_min = int(y_train.min())
            coef = np.around(coef, decimals=3)
            intercept = np.around(intercept, decimals=3)

            for idx, (coef_temp, intercept_temp) in enumerate(zip(coef, intercept), label_min):  # The range of idx is between label_min and label_max.
                terms_temp = [("-" if c < 0 else "+") + " " + str(abs(c)) + f if c != 0 else "" for c, f in zip(coef_temp, features_name)]
                terms_temp_first = (terms_temp[0][2:] if coef_temp[0] > 0 else terms_temp[0]).replace(" ", "")
                formula[f"y (label={idx}):"] = terms_temp_first + " " + " ".join(terms_temp[1:]) + (" - " if intercept_temp < 0 else " + ") + str(abs(intercept_temp))
                print(f"y (label={idx}) = ", formula[f"y (label={idx}):"])

    return formula


# Used by linear models including classification and regression
# ------>

# <------
# Used regresssion and classification models


def plot_permutation_importance(X_test: pd.DataFrame, y_test: pd.DataFrame, trained_model: object, image_config: dict) -> tuple:
    """Plot the permutation Importance.

    Parameters
    ----------
    X_test : pd.DataFrame (n_samples, n_components)
        The testing target values.

    y_test : pd.DataFrame (n_samples, n_components)
    The testing target values.

    trained_model : sklearn algorithm model
        The sklearn algorithm model trained with X_train data.

    image_config : dict
        Image Configuration

    Returns
    -------
    result.importances_mean : ndarray
        The mean of feature importance over repetitions.

    result.importances_std : ndarray
        The standard deviation over repetitions.

    result.importances : ndarray
        The matrix of all feature importance values.
    """

    # create drawing canvas
    fig, ax = plt.subplots(figsize=(image_config["width"], image_config["height"]), dpi=image_config["dpi"])

    columns_name = X_test.columns
    result = permutation_importance(trained_model, X_test, y_test, n_repeats=10, random_state=42, n_jobs=-1)
    sorted_idx = result.importances_mean.argsort()
    ax.boxplot(
        result.importances[sorted_idx].T,
        vert=False,
        labels=np.array(columns_name),
    )

    # automatically optimize picture layout structure
    fig.tight_layout()
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    x_adjustment = (xmax - xmin) * 0.01
    y_adjustment = (ymax - ymin) * 0.01
    ax.axis([xmin - x_adjustment, xmax + x_adjustment, ymin - y_adjustment, ymax + y_adjustment])

    # convert the font of the axes
    x1_label = ax.get_xticklabels()  # adjust the axis label font
    [x1_label_temp.set_fontname(image_config["axislabelfont"]) for x1_label_temp in x1_label]
    y1_label = ax.get_yticklabels()
    [y1_label_temp.set_fontname(image_config["axislabelfont"]) for y1_label_temp in y1_label]

    ax.set_title(
        label=image_config["title_label"],
        fontdict={
            "size": image_config["title_size"],
            "color": image_config["title_color"],
            "family": image_config["title_font"],
        },
        loc=image_config["title_location"],
        pad=image_config["title_pad"],
    )

    return result.importances_mean, result.importances_std, result.importances


# Used regresssion and classification models
# ------>



import os

from dotenv import load_dotenv
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

# Load the .env file
load_dotenv()

# Get the database URL from the environment
database_url = os.getenv("SQLALCHEMY_DATABASE_URL")

# Create the database engine
engine = create_engine(database_url, connect_args={"check_same_thread": False})

# Create a sessionmaker
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

# Create a base class for the models
Base = declarative_base()


def get_db():
    """Get a database session."""
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



import os

import pandas as pd
from auth.dependencies import get_current_active_user
from database import get_db
from fastapi import APIRouter, Depends, HTTPException, UploadFile

from .service import read_all_datasets, read_basic_datasets_info, read_dataset, remove_dataset, upload_dataset

# import subprocess


# from fastapi.responses import PlainTextResponse


# Mock the database
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FAKE_DATABASE_DIR = os.path.join(CURRENT_DIR, "fake_database")

router = APIRouter(
    prefix="/data-mining",
    tags=["data-mining"],
    responses={404: {"description": "Not found"}},
)


@router.post("/upload-dataset")
async def post_dataset(dataset: UploadFile, current_user=Depends(get_current_active_user), db=Depends(get_db)):
    dataset_name = dataset.filename
    data = await dataset.read()
    df = pd.read_excel(data)
    json_df = df.to_json(orient="records")
    db_dataset = upload_dataset(db=db, user_id=current_user.id, dataset_name=dataset_name, json_dataset=json_df)

    os.makedirs(FAKE_DATABASE_DIR, exist_ok=True)
    df.to_excel(os.path.join(FAKE_DATABASE_DIR, "user_data.xlsx"))

    # Return the processed data to the client
    # processed_data = df.to_json(orient="records")
    # print(processed_data)
    # print(type(processed_data))
    # return processed_data

    return {"uploaded_dataset": db_dataset}


@router.delete("/delete-dataset")
async def delete_dataset(dataset_id: int, current_user=Depends(get_current_active_user), db=Depends(get_db)):
    try:
        db_dataset = remove_dataset(db=db, user_id=current_user.id, dataset_id=dataset_id)
    except Exception:
        raise HTTPException(status_code=404, detail="Dataset not found")
    return {"removed_dataset": db_dataset}


@router.get("/get-all-dataset")
async def get_all_datasets(current_user=Depends(get_current_active_user), db=Depends(get_db)):
    return read_all_datasets(db=db, user_id=current_user.id)


@router.get("/basic-datasets-info")
async def get_basic_datasets_info(current_user=Depends(get_current_active_user), db=Depends(get_db)):
    return read_basic_datasets_info(db=db, user_id=current_user.id)


@router.get("/get-dataset")
async def get_dataset(dataset_id: int, current_user=Depends(get_current_active_user), db=Depends(get_db)):
    try:
        db_dataset = read_dataset(db=db, user_id=current_user.id, dataset_id=dataset_id)
    except Exception:
        raise HTTPException(status_code=404, detail="Dataset not found")
    return {"dataset": db_dataset}


# @router.get("/get-raw-data")
# async def get_data():
#     if not os.path.exists(os.path.join(FAKE_DATABASE_DIR, "user_data.xlsx")):
#         return {"message": "No data available"}
#     df = pd.read_excel(os.path.join(FAKE_DATABASE_DIR, "user_data.xlsx"))
#     raw_data = df.to_json(orient="records")
#     print(raw_data)
#     print(type(raw_data))
#     return raw_data


# @router.post("/upload")
# async def upload_data(data: bytes = File(...)):
#     # print(data)
#     print(pd.read_excel(data))
#     # Process the uploaded file as needed
#     # e.g., save it to a specific location or perform data processing
#     return {'message': 'Data uploaded successfully'}


# @router.get("/execute-command")
# async def execute_command(command: str):
#     process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
#     output, _ = process.communicate()
#     return PlainTextResponse(output.decode())



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input, num_input, str_input


def sgd_regression_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters for SGD Regression.

    Returns
    -------
    hyper_parameters : dict
        The hyperparameters.
    """
    print("Loss Function: The loss function to be used. Choose from 'squared_error', 'huber', 'epsilon_insensitive', or 'squared_epsilon_insensitive'.")
    print("Please specify the loss function you want to use. The default is 'squared_error'.")
    losses = ["squared_error", "huber", "epsilon_insensitive", "squared_epsilon_insensitive"]
    loss = str_input(losses, SECTION[2])

    print("Penalty: The penalty (regularization term) to be used. Choose from 'l2', 'l1', 'elasticnet', or None.")
    print("Please specify the penalty (regularization term) you want to use. The default is 'l2'.")
    penalties = ["l2", "l1", "elasticnet", "None"]
    penalty = str_input(penalties, SECTION[2])
    if penalty == "None":
        penalty = None

    l1_ratio = None
    if penalty == "elasticnet":
        print("L1 Ratio: The Elastic Net mixing parameter, with 0 <= l1_ratio <= 1. l1_ratio=0 corresponds to L2 penalty, l1_ratio=1 to L1. The default is 0.15.")
        l1_ratio = float_input(0.15, SECTION[2], "@L1 Ratio: ")

    print("Alpha: The regularization strength. A good starting value could be between 0.0001 and 1, such as 0.01. The default is 0.0001.")
    alpha = float_input(0.0001, SECTION[2], "@Alpha: ")

    print("Fit Intercept: Whether to fit an intercept term. It is generally recommended to set it as True.")
    fit_intercept = bool_input(SECTION[2])

    print("Maximum Number of Iterations: The maximum number of iterations for training. The default is 1000.")
    max_iter = num_input(SECTION[2], "@Max Iterations: ")

    print("Tolerance: The tolerance for stopping criteria. A good starting value could be between 0.000001 and 0.001, such as 0.0001. The default is 0.001.")
    print("If you choose to not set a tolerance value (by pressing Enter without input), training will continue until convergence.")
    tol = float_input(0.001, SECTION[2], "@Tolerance: ")

    print("Shuffle: Whether or not to shuffle the training data after each epoch.")
    shuffle = bool_input(SECTION[2])

    print("Learning Rate: The learning rate schedule. Choose from 'constant', 'optimal', 'invscaling', or 'adaptive'.")
    print("Please indicate the learning rate schedule you want to use. The default is 'invscaling'.")
    learning_rates = ["constant", "optimal", "invscaling", "adaptive"]
    learning_rate = str_input(learning_rates, SECTION[2])

    print("Initial Learning Rate: The initial learning rate for the 'constant', 'invscaling' or 'adaptive' schedules. The default value is 0.01.")
    eta0 = float_input(0.01, SECTION[2], "@Initial Learning Rate: ")

    print("Power T: The exponent for inverse scaling learning rate. The default is 0.25.")
    power_t = float_input(0.25, SECTION[2], "@Power T: ")

    hyper_parameters = {
        "loss": loss,
        "penalty": penalty,
        "alpha": alpha,
        "fit_intercept": fit_intercept,
        "max_iter": max_iter,
        "tol": tol,
        "shuffle": shuffle,
        "learning_rate": learning_rate,
        "eta0": eta0,
        "power_t": power_t,
    }
    if not l1_ratio:
        # Use the default value provided by sklearn.ensemble.SGDRegressor.
        hyper_parameters["l1_ratio"] = 0.15
    else:
        hyper_parameters["l1_ratio"] = l1_ratio
    return hyper_parameters



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import num_input, str_input


def decision_tree_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hype_parameters : dict
    """
    print(
        "Criterion: The function to measure the quality of a split. Supported criteria are “squared_error” for the mean squared error, which is equal to variance reduction"
        " as feature selection criterion and minimizes the L2 loss using the mean of each terminal node, “friedman_mse”, which uses mean squared error"
        " with Friedman’s improvement score for potential splits, “absolute_error” for the mean absolute error, which minimizes the L1 loss using the median of each terminal node, and “poisson”"
        " which uses reduction in Poisson deviance to find splits."
    )
    print("The default value is 'squared_error'. Optional criterions are 'friedman_mse', 'absolute_error' and 'poisson'.")
    criterions = ["squared_error", "friedman_mse", "absolute_error", "poisson"]
    criterion = str_input(criterions, SECTION[2])
    print("Max Depth: The maximum depth of the tree. If None, then nodes are expanded until all leaves are pure or until all leaves contain less than min_samples_split samples.")
    print("Please specify the maximum depth of the tree. A good starting range could be between 3 and 15, such as 4.")
    max_depth = num_input(SECTION[2], "@Max Depth: ")
    print("Min Samples Split: The minimum number of samples required to split an internal node.")
    print("Please specify the minimum number of samples required to split an internal node. A good starting range could be between 2 and 10, such as 3.")
    min_samples_split = num_input(SECTION[2], "@Min Samples Split: ")
    print("Min Samples Leaf: The minimum number of samples required to be at a leaf node.")
    print("Please specify the minimum number of samples required to be at a leaf node. A good starting range could be between 1 and 10, such as 2.")
    min_samples_leaf = num_input(SECTION[2], "@Min Samples Leaf: ")
    print("Max Features: The number of features to consider when looking for the best split.")
    print("Please specify the number of features to consider when looking for the best split. A good starting range could be between 1 and the total number of features in the dataset.")
    max_features = num_input(SECTION[2], "@Max Features: ")
    hyper_parameters = {
        "criterion": criterion,
        "max_depth": max_depth,
        "min_samples_split": min_samples_split,
        "min_samples_leaf": min_samples_leaf,
        "max_features": max_features,
    }
    return hyper_parameters



from database import Base
from sqlalchemy import Boolean, Column, Integer, String
from sqlalchemy.orm import relationship


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True, index=True)
    username = Column(String, unique=True, index=True)
    email = Column(String, unique=True, index=True)
    hashed_password = Column(String)
    upload_count = Column(Integer, default=0)
    is_active = Column(Boolean, default=True)

    datasets = relationship("Dataset", back_populates="user")



from enum import Enum


class ModeOption(Enum):
    REGRESSION = "Regression"
    CLASSIFICATION = "Classification"
    CLUSTERING = "Clustering"
    DIMENSIONAL_REDUCTION = "Dimensional Reduction"
    ABNORMAL_DETECTION = "Abnormal Detection"


class ModeOptionWithMissingValues(Enum):
    REGRESSION = "Regression"
    CLASSIFICATION = "Classification"
    CLUSTERING = "Clustering"


# -*- coding: utf-8 -*-
from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import bool_input, float_input, num_input


def ridge_regression_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("Alpha: This hyperparameter represents the coefficient of the norm, which controls the degree of contraction constraint.")
    print("Please indicate the coefficient of alpha. A good starting range could be between 0.001 and 2, such as 1.")
    alpha = float_input(0.01, SECTION[2], "@Alpha: ")
    print("Fit Intercept: This hyperparameter represents whether the model is evaluated with constant terms.")
    print("Please indicate whether there is a parameter entry. It is generally recommended to leave it set to True.")
    fit_intercept = bool_input(SECTION[2])
    print("Max Iter: This hyperparameter represents the maximum number of iterations for the solver to converge.")
    print("Please indicate the maximum number of iterations. A good starting range could be between 1000 and 10000, such as 1000.")
    max_iter = num_input(SECTION[2], "@Max Iter: ")
    print("Tolerance: This hyperparameter represents the tolerance of the optimization method.")
    print("Please indicate the tolerance. A good starting range could be between 0.0001 and 0.001, such as 0.0001.")
    tol = float_input(0.0001, SECTION[2], "@Tolerance: ")
    # print("Selection: This hyperparameter represents the method of selecting the regularization coefficient.")
    # print("Please indicate the method of selecting the regularization coefficient. It is generally recommended to leave it set to 'cyclic'.")
    # selections = ["cyclic", "random"]
    # selection = str_input(selections, SECTION[2])
    hyper_parameters = {
        "alpha": alpha,
        "fit_intercept": fit_intercept,
        "max_iter": max_iter,
        "tol": tol,
    }
    return hyper_parameters



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

import sphinx_rtd_theme

project = "Geochemistry π"
copyright = "2023, ZJUEarthData"
author = "ZJUEarthData"
# release = 'v0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# extensions = []
extensions = ["sphinxcontrib.napoleon", "sphinxcontrib.apidoc", "sphinx.ext.viewcode", "sphinx.ext.todo", "m2r2"]

templates_path = ["_templates"]
exclude_patterns = []

language = "EN"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


html_theme = "furo"
# html_theme = 'classic'
# html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ["_static"]

html_theme_options = {
    "top_of_page_button": "edit",
    "sidebar_hide_name": True,
    "navigation_with_keys": True,
    "announcement": "<em>More tutorial videos can be found on bilibili: ZJU_Earth_Data</em>",
    "light_css_variables": {
        "color-brand-primary": "blue",
        # "color-brand-content": "brown",
        "color-admonition-background": "orange",
        "font-stack": "Times New Roman, Times, serif",
    },
}


html_logo = "https://user-images.githubusercontent.com/66779478/239791119-3b1fe8c9-5f99-49f5-aa31-edf5a683372b.png"

source_suffix = [".rst", ".md"]


project_path = "../../geochemistrypi"
# autodoc_mock_imports = ["geochemistrypi"]
sys.path.insert(0, os.path.abspath("../.."))
# sys.path.insert(0, os.path.abspath(project_path))
# print(os.path.abspath(project_path))
# sys.path.insert(1, os.path.abspath("../../geochemistrypi/"))
# sys.path.insert(2, os.path.abspath("../../geochemistrypi/data_mining/"))
# sys.path.insert(3, os.path.abspath(".."))
# sys.path.append("../geochemistrypy/geochemistrypy")
# ...

apidoc_module_dir = project_path
apidoc_output_dir = "python_apis"

apidoc_separate_modules = False

# Disable displaying the module name as a prefix to class and function names
add_module_names = False

# Disable displaying the module name as a prefix to class and function names
# autodoc_default_flags = ['package', 'undoc-members', 'private-members', 'special-members', 'inherited-members', 'show-inheritance', 'noindex']
html_show_sourcelink = False
html_use_smartypants = True
html_split_index = False



import matplotlib.pyplot as plt
import numpy as np

from ..constants import GEO_IMAGE_PATH
from ..utils.base import save_fig


# TODO(Sany hecan@mail2.sysu.edu.cn): test and incorporate it into the current framework
def elements_ratio_map(col1, col2, df):
    """plot the ratio of two different chemical elements

    :param col1: pd.Series, one column in the data set represents one specific elements
    :param col2: pd.Series, one column in the data set represents one specific elements
    :param df: pd.DataFrame, the data set
    """
    # Used to display the minus sign normally
    plt.rcParams["axes.unicode_minus"] = False

    fig = plt.figure(dpi=360)
    ax = fig.add_subplot(111)

    # Data processing
    data = df.loc[:, [col1, col2]]
    data = data.reindex(columns=["add", col1, col2], fill_value=1)
    x = data.loc[:, [col1]]
    y = data.loc[:, [col2]]
    xarray = np.array(x)
    yarray = np.array(y)
    xmax = np.max(xarray)
    xmin = np.min(xarray)
    ymin = np.min(yarray)
    ymax = np.max(yarray)

    # Use regression to fit data
    xMat = np.mat(data.iloc[:, :-1].values)
    yMat = np.mat(data.iloc[:, -1].values).T
    xTx = xMat.T * xMat
    if np.linalg.det(xTx) == 0:
        print("The matrix is singular and cannot be inverted!")
        return
    ws = xTx.I * (xMat.T * yMat)
    yHat = xMat * ws
    plt.plot(xMat[:, 1], yHat, c="black", ls="-")
    ws = ws.tolist()
    plt.scatter(0, 0, marker="o", color="r", edgecolors="r", s=0.1, alpha=0.6, label="Positive")
    plt.scatter(0, 0, marker="^", color="b", edgecolors="b", s=0.1, alpha=0.6, label="Negative")
    plt.scatter(0, 0, color="gray", edgecolors="gray", alpha=0.01, s=1, lw=1, label=("Major elements(n=" + str(len(x)) + ")"))

    for index in range(len(x)):

        n = int(index)
        y = ws[1] * xarray[n] + ws[0]
        m = y * xarray[n]
        k = xarray[n] * yarray[n]

        if k > m:
            plt.scatter(xarray[n], yarray[n], marker="o", color="r", edgecolors="r", s=10, lw=2, alpha=0.6)
        else:
            plt.scatter(xarray[n], yarray[n], marker="^", color="b", edgecolors="b", s=10, alpha=0.3)

    plt.xlabel(col1, size=15)
    plt.ylabel(col2, size=15)
    plt.xlim(0.96 * xmin, 1.02 * xmax)
    plt.ylim(ymin, ymax)
    plt.legend(edgecolor="gray", facecolor="gray", fontsize="large", loc="upper right", markerscale=8.5)
    plt.text(0.011, 0.004, str(col2) + "/" + str(col1) + "=" + str(ws[1]), transform=ax.transAxes)
    save_fig(f"elements ratio - {col1.name} vs {col2.name}", GEO_IMAGE_PATH)



# -*- coding: utf-8 -*-

from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from rich import print
from sklearn.ensemble import IsolationForest

from ..utils.base import clear_output
from ._base import WorkflowBase
from .func.algo_abnormaldetection._iforest import isolation_forest_manual_hyper_parameters


class AbnormalDetectionWorkflowBase(WorkflowBase):
    """The base workflow class of abnormal detection algorithms."""

    # common_function = []

    def __init__(self) -> None:
        super().__init__()
        self.mode = "Abnormal Detection"

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> None:
        """Fit the model by Scikit-learn framework."""
        self.X = X
        self.model.fit(X)

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        """Perform Abnormal Detection on samples in X by Scikit-learn framework."""
        y_predict = self.model.predict(X)
        return y_predict

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        return dict()

    @staticmethod
    def _detect_data(X: pd.DataFrame, detect_label: np.ndarray) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """Merge the detection results into the source data.

        Parameters
        ----------
        X : pd.DataFrame
            The original data.

        detect_label : np.ndarray
            The detection labels for each data point.

        Returns
        -------
        X_abnormal_detection : pd.DataFrame
            DataFrame containing the original data with detection results.

        X_normal : pd.DataFrame
            DataFrame containing the normal data points.

        X_abnormal : pd.DataFrame
            DataFrame containing the abnormal data points.
        """
        X_abnormal_detection = X.copy()
        # Merge detection results into the source data
        X_abnormal_detection["is_abnormal"] = detect_label
        X_normal = X_abnormal_detection[X_abnormal_detection["is_abnormal"] == 1]
        X_abnormal = X_abnormal_detection[X_abnormal_detection["is_abnormal"] == -1]

        return X_abnormal_detection, X_normal, X_abnormal

    def common_components(self) -> None:
        """Invoke all common application functions for abnormal detection algorithms by Scikit-learn framework."""
        pass


class IsolationForestAbnormalDetection(AbnormalDetectionWorkflowBase):
    """The automation workflow of using Isolation Forest algorithm to make insightful products."""

    name = "Isolation Forest"
    # special_function = []

    def __init__(
        self,
        n_estimators: int = 100,
        max_samples: Union[str, int, float] = "auto",
        contamination: Union[str, float] = "auto",
        max_features: Union[int, float] = 1.0,
        bootstrap: bool = False,
        n_jobs: Optional[int] = None,
        random_state: Optional[int] = None,
        verbose: int = 0,
        warm_start: bool = False,
    ) -> None:
        """
        Isolation Forest Algorithm.

        Return the anomaly score of each sample using the IsolationForest algorithm

        The IsolationForest 'isolates' observations by randomly selecting a feature
        and then randomly selecting a split value between the maximum and minimum
        values of the selected feature.

        Since recursive partitioning can be represented by a tree structure, the
        number of splittings required to isolate a sample is equivalent to the path
        length from the root node to the terminating node.

        This path length, averaged over a forest of such random trees, is a
        measure of normality and our decision function.

        Random partitioning produces noticeably shorter paths for anomalies.
        Hence, when a forest of random trees collectively produce shorter path
        lengths for particular samples, they are highly likely to be anomalies.

        Read more in the :ref:`User Guide <isolation_forest>`.

        .. versionadded:: 0.18

        Parameters
        ----------
        n_estimators : int, default=100
            The number of base estimators in the ensemble.

        max_samples : "auto", int or float, default="auto"
            The number of samples to draw from X to train each base estimator.
                - If int, then draw `max_samples` samples.
                - If float, then draw `max_samples * X.shape[0]` samples.
                - If "auto", then `max_samples=min(256, n_samples)`.

            If max_samples is larger than the number of samples provided,
            all samples will be used for all trees (no sampling).

        contamination : 'auto' or float, default='auto'
            The amount of contamination of the data set, i.e. the proportion
            of outliers in the data set. Used when fitting to define the threshold
            on the scores of the samples.

                - If 'auto', the threshold is determined as in the
                original paper.
                - If float, the contamination should be in the range (0, 0.5].

            .. versionchanged:: 0.22
            The default value of ``contamination`` changed from 0.1
            to ``'auto'``.

        max_features : int or float, default=1.0
            The number of features to draw from X to train each base estimator.

                - If int, then draw `max_features` features.
                - If float, then draw `max(1, int(max_features * n_features_in_))` features.

            Note: using a float number less than 1.0 or integer less than number of
            features will enable feature subsampling and leads to a longer runtime.

        bootstrap : bool, default=False
            If True, individual trees are fit on random subsets of the training
            data sampled with replacement. If False, sampling without replacement
            is performed.

        n_jobs : int, default=None
            The number of jobs to run in parallel for both :meth:`fit` and
            :meth:`predict`. ``None`` means 1 unless in a
            :obj:`joblib.parallel_backend` context. ``-1`` means using all
            processors. See :term:`Glossary <n_jobs>` for more details.

        random_state : int, RandomState instance or None, default=None
            Controls the pseudo-randomness of the selection of the feature
            and split values for each branching step and each tree in the forest.

            Pass an int for reproducible results across multiple function calls.
            See :term:`Glossary <random_state>`.

        verbose : int, default=0
            Controls the verbosity of the tree building process.

        warm_start : bool, default=False
            When set to ``True``, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit a whole
            new forest. See :term:`the Glossary <warm_start>`.

            .. versionadded:: 0.21

        References
        ----------
        Scikit-learn API: sklearn.ensemble.IsolationForest
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html#
        """

        super().__init__()
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.contamination = contamination
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.warm_start = warm_start

        if random_state:
            self.random_state = random_state

        # If 'random_state' is None, 'self.random_state' comes from the parent class 'WorkflowBase'
        self.model = IsolationForest(
            n_estimators=self.n_estimators,
            max_samples=self.max_samples,
            contamination=self.contamination,
            max_features=self.max_features,
            bootstrap=self.bootstrap,
            n_jobs=self.n_jobs,
            random_state=self.random_state,
            verbose=self.verbose,
            warm_start=self.warm_start,
        )

        self.naming = IsolationForestAbnormalDetection.name

    @classmethod
    def manual_hyper_parameters(cls) -> Dict:
        """Manual hyper-parameters specification."""
        print(f"-*-*- {cls.name} - Hyper-parameters Specification -*-*-")
        hyper_parameters = isolation_forest_manual_hyper_parameters()
        clear_output()
        return hyper_parameters

    def special_components(self, **kwargs) -> None:
        """Invoke all special application functions for this algorithms by Scikit-learn framework."""
        pass



# -*- coding: utf-8 -*-
from itertools import cycle

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import kde


def plot_2d_scatter_diagram(data: pd.DataFrame, algorithm_name: str) -> None:
    """
    Plot a 2D scatter diagram for dimensionality reduction results.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
        Data after dimensionality reduction.

    pc : pd.DataFrame (n_features, n_components)
        principal components.

    algorithm_name : str
        The name of the dimensionality reduction algorithm.

        labels : List[str]
        The type of tag of the samples in the data set.
    """
    markers = ["+", "v", ".", "d", "o", "s", "1", "D", "X", "^", "p", "<", "*", "H", "3", "P"]
    colors = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#33a02c",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
    ]

    marker_cycle = cycle(markers)
    color_cycle = cycle(colors)

    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111)

    # Plot the data
    for i, label in enumerate(data.index):
        colors = next(color_cycle)
        markers = next(marker_cycle)
        ax.scatter(data.iloc[:, 0], data.iloc[:, 1], c=colors, marker=markers, label=label)

    ax.set_xlabel("Component 1")
    ax.set_ylabel("Component 2")
    ax.set_title(f"{algorithm_name} Dimensionality Reduction Results")
    ax.legend(loc="upper right")

    plt.grid(True)


def plot_heatmap(data: pd.DataFrame, algorithm_name: str) -> None:
    """
    Plot a heatmap for dimensionality reduction results.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
        Data after dimensionality reduction.

    algorithm_name : str
        The name of the dimensionality reduction algorithm.
    """
    plt.figure(figsize=(10, 8))
    sns.heatmap(data, cmap="viridis")
    plt.title(f"{algorithm_name} Dimensionality Reduction Heatmap")
    plt.xlabel("Component")
    plt.ylabel("Sample")


def plot_contour(data: pd.DataFrame, algorithm_name: str) -> None:
    """
    Plot a contour plot for dimensionality reduction results.

    Parameters
    ----------
    data : pd.DataFrame (n_samples, n_components)
        Data after dimensionality reduction.

    algorithm_name : str
        The name of the dimensionality reduction algorithm.
    """
    # Calculate the density
    x = data.iloc[:, 0]
    y = data.iloc[:, 1]
    buffer = max(x.max() - x.min(), y.max() - y.min()) * 0.05
    xmin, xmax = x.min() - buffer, x.max() + buffer
    ymin, ymax = y.min() - buffer, y.max() + buffer

    xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    positions = np.vstack([xx.ravel(), yy.ravel()])
    values = np.vstack([x, y])
    kernel = kde.gaussian_kde(values)
    f = np.reshape(kernel(positions).T, xx.shape)

    # Plot the contour
    plt.figure(figsize=(10, 8))
    plt.contourf(xx, yy, f, cmap="viridis", alpha=0.5)
    plt.colorbar(label="Density")
    plt.scatter(x, y, marker="o", color="black", alpha=0.5)
    plt.xlabel(f"{data.columns[0]}")
    plt.ylabel(f"{data.columns[1]}")
    plt.title(f"{algorithm_name} Dimensionality Reduction Contour Plot")
    plt.grid(True)



# -*- coding: utf-8 -*-
from typing import List

import numpy as np
import pandas as pd
from rich import print
from sklearn.feature_selection import GenericUnivariateSelect, SelectKBest, f_classif, f_regression
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from .data_readiness import show_data_columns


def feature_scaler(X: pd.DataFrame, method: List[str], method_idx: int) -> tuple[dict, np.ndarray]:
    """Apply feature scaling methods.

    Parameters
    ----------
    X : pd.DataFrame
        The dataset.

    method : str
        The feature scaling methods.

    method_idx : int
        The index of methods.

    Returns
    -------
    feature_scaling_config : dict
        The feature scaling configuration.

    X_scaled : np.ndarray
        The dataset after imputing.
    """
    if method[method_idx] == "Min-max Scaling":
        scaler = MinMaxScaler()
    elif method[method_idx] == "Standardization":
        scaler = StandardScaler()
    try:
        X_scaled = scaler.fit_transform(X)
    except ValueError:
        print("The selected feature scaling method is not applicable to the dataset!")
        print("Please check the dataset to find the reason.")
    feature_scaling_config = {type(scaler).__name__: scaler.get_params()}
    return feature_scaling_config, X_scaled


def feature_selector(X: pd.DataFrame, y: pd.DataFrame, feature_selection_task: int, method: List[str], method_idx: int) -> tuple[dict, pd.DataFrame]:
    """Apply feature selection methods.

    Parameters
    ----------
    X : pd.DataFrame
        The feature dataset.

    y : pd.DataFrame
        The label dataset.

    feature_selection_task : int
        Feature selection for regression or classification tasks.

    method : str
        The feature selection methods.

    method_idx : int
        The index of methods.

    Returns
    -------
    feature_selection_config : dict
        The feature selection configuration.

    X_selected : pd.DataFrame
        The feature dataset after selecting.
    """
    print("-- Original Features --")
    show_data_columns(X.columns)

    features_num = len(X.columns)
    print(f"The original number of features is {features_num}, and your input must be less than {features_num}.")
    features_retain_num = int(input("Please enter the number of features to retain.\n" "@input: "))

    if feature_selection_task == 1:
        score_func = f_regression
    elif feature_selection_task == 2:
        score_func = f_classif

    if method[method_idx] == "Generic Univariate Select":
        selector = GenericUnivariateSelect(score_func=score_func, mode="k_best", param=features_retain_num)
    elif method[method_idx] == "Select K Best":
        selector = SelectKBest(score_func=score_func, k=features_retain_num)

    try:
        selector.fit(X, y)
        features_selected = selector.get_feature_names_out()
        X = X[features_selected]
    except ValueError:
        print("The selected feature selection method is not applicable to the dataset!")
        print("Please check the dataset to find the reason.")

    feature_selection_config = {type(selector).__name__: selector.get_params()}
    return feature_selection_config, X



# -*- coding: utf-8 -*-
import pytest
from data.data_readiness import limit_num_input, num2option, num_input
from pytest import MonkeyPatch


def test_num2option() -> None:
    array = ["a", "b", "c"]
    num2option(array)


def test_empty_num2option() -> None:
    array = []
    num2option(array)


def test_valid_num_input_one_time(monkeypatch: MonkeyPatch) -> None:
    inputs = ["1"]
    prefix = "Data"
    slogan = "@Number"
    monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
    assert num_input(prefix, slogan) == 1


def test_invalid_num_input_two_time(monkeypatch: MonkeyPatch) -> None:
    inputs = ["string", "1"]
    prefix = "Data"
    slogan = "@Number"
    # when the input is string type, then move on until the function get a int value
    monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
    assert num_input(prefix, slogan) == 1


def test_invalid_num_input_equal(monkeypatch: MonkeyPatch) -> None:
    with pytest.raises(AssertionError):
        inputs = ["2"]
        prefix = "Data"
        slogan = "@Number"
        monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
        assert num_input(prefix, slogan) == 1


# @pytest.fixture
# def input_func(monkeypatch: MonkeyPatch) -> num_input:
#     inputs = ['3']
#     prefix = 'Data'
#     slogan = '@Number'
#     monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
#     return num_input(prefix, slogan)


# def test_valid_limit_num_input(input_func: num_input) -> None:
#     option_list = ['Option1', 'Option2', 'Option3']
#     prefix = 'Data'
#     assert limit_num_input(option_list, prefix, input_func) == 3


def test_valid_limit_num_input_one_time(monkeypatch: MonkeyPatch) -> None:
    option_list = ["Option1", "Option2", "Option3"]
    inputs = ["3"]
    prefix = "Data"
    monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
    assert limit_num_input(option_list, prefix, num_input) == 3


def test_valid_limit_num_input_two_time(monkeypatch: MonkeyPatch) -> None:
    option_list = ["Option1", "Option2", "Option3"]
    inputs = ["4", "1"]
    prefix = "Data"
    monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
    assert limit_num_input(option_list, prefix, num_input) == 1


def test_invalid_limit_num_input(monkeypatch: MonkeyPatch) -> None:
    with pytest.raises(AssertionError):
        option_list = ["Option1", "Option2", "Option3"]
        inputs = ["4", "1"]
        prefix = "Data"
        monkeypatch.setattr("builtins.input", lambda _: inputs.pop(0))
        assert limit_num_input(option_list, prefix, num_input) == 4



from typing import Dict

from rich import print

from ....constants import SECTION
from ....data.data_readiness import num_input, str_input


def agglomerative_manual_hyper_parameters() -> Dict:
    """Manually set hyperparameters.

    Returns
    -------
    hyper_parameters : dict
    """
    print("N Clusters: The number of clusters to form as well as the number of centroids to generate.")
    print("Please specify the number of clusters for agglomerative. A good starting range could be between 2 and 10, such as '4'.")
    n_clusters = num_input(SECTION[2], "N Clusters: ")
    print("linkage: The linkage criterion determines which distance to use between sets of observation. ")
    print("Please specify the linkage criterion. It is generally recommended to leave it set to 'ward'.")
    linkages = ["ward", "complete", "average", "single"]
    linkage = str_input(linkages, SECTION[2])
    hyper_parameters = {
        "n_clusters": n_clusters,
        "linkage": linkage,
    }
    return hyper_parameters



from abc import ABCMeta, abstractmethod


class ModelSelectionBase(metaclass=ABCMeta):
    """Abstract base class for model selection"""

    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        self.workflow = None
        self.transformer_config = {}

    @abstractmethod
    def activate(self, *args, **kwargs) -> None:
        pass



import os
import sys
import time
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import openpyxl.utils.exceptions
import pandas as pd
from rich import print
from sklearn.model_selection import train_test_split

from ..constants import BUILT_IN_DATASET_PATH

# from utils.exceptions import InvalidFileError


def read_data(file_path: Optional[str] = None, is_own_data: int = 2, prefix: Optional[str] = None, slogan: Optional[str] = "@File: "):
    """Read the data set.

    Parameters
    ----------
    file_path : str, optional
        The path of the data set, by default None

    is_own_data : int, default=2
        1: own data set; 2: built-in data set

    prefix : str, optional
        The prefix of the data set, by default None

    slogan : str, optional
        The slogan of the data set, by default "@File: "

    Returns
    -------
    pd.DataFrame
        The data set read
    """
    if is_own_data == 1:
        data_path = file_path
    else:
        data_path = os.path.join(BUILT_IN_DATASET_PATH, file_path)
    try:
        if data_path.endswith(".xlsx"):
            data = pd.read_excel(data_path, engine="openpyxl")
        elif data_path.endswith(".csv"):
            data = pd.read_csv(data_path)
        return data
    except ImportError as err:
        print(err)
        print("[red]Warning: on Mac, input the following command in terminal: pip3 install openpyxl[red]")
        raise err
    except FileNotFoundError as err:
        print(err)
        print("[red]Warning: please put your own data in the right place and input the completed data set name including" " the stored path and suffix[red]")
        raise err
    except openpyxl.utils.exceptions.InvalidFileException as err:
        print(err)
        print("[red]Warning: please put your own data in the right place and input the completed data set name including" " the stored path and suffix[red]")
        raise err
    except Exception:
        print(f"[red]Unexpected error: {sys.exc_info()[0]} - check the last line of Traceback about the error information[red]")
        raise Exception


def basic_info(data: pd.DataFrame) -> None:
    """Show the basic information of the data set.

    Parameters
    ----------
    data : pd.DataFrame
        The data set to be shown.
    """
    print(data.info())


def show_data_columns(columns_name: pd.Index, columns_index: Optional[List] = None) -> None:
    """Show the column names of the data set.

    Parameters
    ----------
    columns_name : pd.Index
        The column names of the data set.

    columns_index : list, default=None
        The column index of the data set.
    """
    print("-" * 20)
    print("Index - Column Name")
    if columns_index is None:
        for i, j in enumerate(columns_name):
            print(i + 1, "-", j)
    else:
        # specify the designated column index
        for i, j in zip(columns_index, columns_name):
            print(i + 1, "-", j)
    print("-" * 20)


def select_columns(columns_range: Optional[str] = None) -> List[int]:
    """Select the columns of the data set.

    Parameters
    ----------
    columns_range : str, default=None
        The columns range of the data set.

    Returns
    -------
    list
        The columns selected.
    """
    columns_selected = []
    temp = columns_range.split(";")
    for i in range(len(temp)):
        if isinstance(eval(temp[i]), int):
            columns_selected.append(eval(temp[i]))
        else:
            min_max = eval(temp[i])
            index1 = min_max[0]
            index2 = min_max[1]
            j = [index2 - j for j in range(index2 - index1 + 1)]
            columns_selected = columns_selected + j

    # delete the repetitive index in the list
    columns_selected = list(set(columns_selected))
    # sort the index list
    columns_selected.sort()
    # reindex by subtracting 1 due to python list traits
    columns_selected = [columns_selected[i] - 1 for i in range(len(columns_selected))]
    return columns_selected


def create_sub_data_set(data: pd.DataFrame, allow_empty_columns: bool = False) -> pd.DataFrame:
    """Create a sub data set.

    Parameters
    ----------
    data : pd.DataFrame
        The data set to be processed.

    allow_empty_columns : bool, optional
        Whether to include empty columns in the sub data set. The default is False.

    Returns
    -------
    pd.DataFrame
        The sub data set.
    """
    sub_data_set_columns_range = str(
        input(
            "Select the data range you want to process.\n"
            "Input format:\n"
            'Format 1: "[**, **]; **; [**, **]", such as "[1, 3]; 7; [10, 13]" '
            "--> you want to deal with the columns 1, 2, 3, 7, 10, 11, 12, 13 \n"
            'Format 2: "**", such as "7" --> you want to deal with the columns 7 \n'
            "@input: "
        )
    )
    while True:
        if ("【" in sub_data_set_columns_range) or ("】" in sub_data_set_columns_range):
            print("There is a problem with the format of the parentheses entered !")
            time.sleep(0.5)
            sub_data_set_columns_range = str(input("-----* Please enter again *-----\n@input: "))
            judge = True
        else:
            monitor_number = 0
            for i in ["[", "]"]:
                if i in sub_data_set_columns_range:
                    monitor_number = monitor_number + 1
            if monitor_number % 2 != 0:
                print("There is a problem with the format of the parentheses entered !")
                time.sleep(0.5)
                sub_data_set_columns_range = str(input("-----* Please enter again *-----\n@input: "))
                judge = True
            sub_data_set_columns_range = sub_data_set_columns_range.replace(" ", "")
            temp = sub_data_set_columns_range.split(";")
            if len(sub_data_set_columns_range) != 0:
                for i in range(len(temp)):
                    if isinstance(eval(temp[i]), int):
                        if int(temp[i]) > int(data.shape[1]):
                            print("The input {} is incorrect!".format(temp[i]))
                            print("The number you entered is out of the range of options: 1 - {}".format(data.shape[1]))
                            time.sleep(0.5)
                            sub_data_set_columns_range = input("-----* Please enter again *-----\n@input: ")
                            judge = True
                            break
                        else:
                            judge = False
                    else:
                        min_max = eval(temp[i])
                        if int(min_max[0]) >= int(min_max[1]):
                            print("There is a problem with the format of the data you entered!")
                            time.sleep(0.5)
                            sub_data_set_columns_range = input("-----* Please enter again *-----\n@input: ")
                            judge = True
                            break
                        elif int(min_max[1]) > int(data.shape[1]):
                            print("The input {} is incorrect!".format(temp[i]))
                            print("The number you entered is out of the range of options: 1 - {}".format(data.shape[1]))
                            time.sleep(0.5)
                            sub_data_set_columns_range = input("-----* Please enter again *-----\n@input: ")
                            judge = True
                            break
                        else:
                            judge = False
            else:
                print("You have not entered the sequence number of the selected data!")
                print("The number you entered should be in the range of options: 1 - {}".format(data.shape[1]))
                time.sleep(0.5)
                sub_data_set_columns_range = input("-----* Please enter again *-----\n@input: ")
                judge = True

        if judge is False:
            break

    while True:
        try:
            # column name
            sub_data_set_columns_selected = select_columns(sub_data_set_columns_range)
            judge = False
        except SyntaxError:
            print("Warning: Please use English input method editor.")
            judge = True
            sub_data_set_columns_range = str(input("@input: "))
        except NameError:
            print("Warning: Please follow the rules and re-enter.")
            judge = True
            sub_data_set_columns_range = str(input("@input: "))
        except UnicodeDecodeError:
            print("Warning: Please use English input method editor.")
            judge = True
            sub_data_set_columns_range = str(input("@input: "))
        except IndexError:
            print("Warning: Please follow the rules and re-enter.")
            judge = True
            sub_data_set_columns_range = str(input("@input: "))
        except TypeError:
            print("Warning: Please follow the rules and re-enter.")
            judge = True
            sub_data_set_columns_range = str(input("@input: "))
        else:
            data_checking = data.iloc[:, sub_data_set_columns_selected]
            for i in data_checking.columns.values:
                df_test = pd.DataFrame(data_checking[i])
                test_columns = df_test.columns
                v_value = int(df_test.isnull().sum())
                if not allow_empty_columns and v_value == len(df_test):
                    print(f"Warning: The selected column {df_test.columns.values} is an empty column! It will be automatically removed.")
                if df_test[test_columns[0]].dtype in ["int64", "float64"]:
                    continue
                else:
                    print(f"Warning: The data type of selected column {df_test.columns.values} is not numeric!" " Please make sure that the selected data type is numeric and re-enter.")
                    judge = True
            if judge is True:
                sub_data_set_columns_range = str(input("@input: "))
        if judge is False:
            break

    # select designated column
    sub_data_set = data.iloc[:, sub_data_set_columns_selected]
    if not allow_empty_columns:
        sub_data_set = sub_data_set.dropna(axis=1, how="all")
    show_data_columns(sub_data_set.columns, sub_data_set_columns_selected)
    return sub_data_set


def data_split(X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series], test_size: float = 0.2) -> Dict:
    """Split arrays or matrices into random train and test subsets.

    Parameters
    ----------
    X : pd.DataFrame
        The data to be split.

    y : pd.DataFrame or pd.Series
        The target variable to be split.

    test_size : float, default=0.2
        Represents the proportion of the dataset to include in the test split.

    Returns
    -------
    dict
        A dictionary containing the split data.
    """
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=42)
    return {"X Train": X_train, "X Test": X_test, "Y Train": y_train, "Y Test": y_test}


def num2option(items: List[str]) -> None:
    """List all the options serially.

    Parameters
    ----------
    items : list
        a series of items need to be enumerated
    """
    for i, j in enumerate(items):
        print(str(i + 1) + " - " + j)


def num_input(prefix: Optional[str] = None, slogan: Optional[str] = "@Number: ") -> int:
    """Get the number of the desired option.

    Parameters
    ----------
    prefix : str, default=None
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    slogan : str, default="@Number: "
        It acts like the first parameter of input function in Python, which output the hint.

    Returns
    -------
    option: int
        An option number. Be careful that 'option = real index  + 1'
    """
    # capture exception: input is not digit
    while True:
        option = input(f"({prefix}) ➜ {slogan}").strip()
        if option.isdigit():
            option = int(option)
            if isinstance(option, int):
                break
        else:
            print("Caution: The input is not a positive integer number. Please input the right number again!")
    return option


def np2pd(array: np.ndarray, columns_name: List[str]) -> pd.DataFrame:
    """Convert numpy array to pandas dataframe.

    Parameters
    ----------
    array : np.ndarray
        The numpy array to be converted.

    columns_name : List[str]
        The column names of the dataframe.

    Returns
    -------
    pd.DataFrame
        The converted dataframe.
    """
    return pd.DataFrame(array, columns=columns_name)


def limit_num_input(option_list: List[str], prefix: str, input_func: num_input) -> int:
    """Limit the scope of the option.

    Parameters
    ----------
    option_list : List[str]
        All the options provided are stored in a list.

    prefix : str
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    input_func: function
        The function of input_func.

    Returns
    -------
    option: int
        An option number. Be careful that 'option = real index  + 1'
    """
    while True:
        # in case that the option number is beyond the maximum
        option = input_func(prefix)
        if option not in range(1, len(option_list) + 1):
            print("Caution: The number is invalid. Please enter the correct number inside the scope!")
        else:
            break
    return option


def float_input(default: float, prefix: Optional[str] = None, slogan: Optional[str] = "@Number: ") -> float:
    """Get the number of the desired option.

    Parameters
    ----------
    default: float
         If the user does not enter anything, it is assigned to option.

    prefix : str, default=None
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    slogan : str, default="@Number: "
        It acts like the first parameter of input function in Python, which output the hint.

    Returns
    -------
    option: float or int
        An option number.
    """
    while True:
        option = input(f"({prefix}) ➜ {slogan}").strip()
        if option.isdigit() or option.replace(".", "").isdigit():
            option = float(option)
            break
        elif len(option) == 0:
            option = default
            break
        else:
            print("Caution: The input is not a positive number. Please input the right number again!")
    return option


def str_input(option_list: List[str], prefix: Optional[str] = None) -> str:
    """Get the string of the desired option.

    Parameters
    ----------
    option_list : list
        All the options provided are stored in a list.

    prefix : str, default=None
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    Returns
    -------
    option: str
        A string of the desired option.
    """
    num2option(option_list)
    option_num = limit_num_input(option_list, prefix, num_input)
    option = option_list[option_num - 1]
    return option


def bool_input(prefix: Optional[str] = None) -> bool:
    """Get the number of the desired option.

    Parameters
    ----------
    prefix : str, default=None
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    Returns
    -------
    bool
        A boolean value.
    """
    bool_value = ["True", "False"]
    option = str_input(bool_value, prefix)
    return True if option == "True" else False


def tuple_input(default: Tuple[int], prefix: Optional[str] = None, slogan: Optional[str] = None) -> Tuple[int]:
    """Get the tuple of the desired option.

    Parameters
    ----------
    default: Tuple[int]
         If the user does not enter anything, it is assigned to option.

    prefix : str, default=None
        It indicates which section the user currently is in on the UML, which is shown on the command-line console.

    slogan : str, default=None
        It acts like the first parameter of input function in Python, which output the hint.

    Returns
    -------
    option: tuple
        A numeric tuple.
    """
    while True:
        option = input(
            "Determine the architecture of the multi-layer perceptron.\n"
            "Input format:\n"
            'Format 1: "(**,)", such as "(100,)"\n'
            "--> You want to set one hidden layer with 100 neurons for the multi-layer perceptron.\n"
            'Format 2: "(**, **)", such as "(50, 25)"\n'
            "--> You want to set two hidden layers in order with 50 neurons and 25 neurons respectively"
            " for the multi-layer perceptron.\n"
            'Format 3: "(**, **, **)", such as "(64, 32, 8)"\n'
            "--> You want to set three hidden layers in order 64 neurons, 32 neurons and 8 neurons"
            " respectively for the multi-layer perceptron.\n"
            f"({prefix}) ➜ {slogan}"
        ).strip()
        if len(option) == 0:
            option = default
            break
        else:
            option = eval(option)
            break
    return option
